/***************************************************************************
 *
 * Copyright (c) 2000 BalaBit IT Ltd, Budapest, Hungary
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Id: pysatyr.c,v 1.7 2000/10/10 18:25:58 bazsi Exp $
 *
 * Author  : Bazsi
 * Auditor : kisza
 * Last version: 1.1
 * Notes:
 *   - comments
 *   - typed functions without return
 *
 * Author  : Bazsi
 * Auditor : kisza
 * Last version: 1.1
 * Notes:
 *   - comments
 *   - typed functions without return
 *
 ***************************************************************************/

#include <zorp/pysatyr.h>
#include <zorp/pysockaddr.h>
#include <zorp/pylabelset.h>
#include <zorp/log.h>

extern PyTypeObject z_py_zorp_satyr_type;

static PyObject *
z_py_zorp_satyr_authenticator_new_instance(PyObject *s, PyObject *args)
{
  ZorpAuthConnection *authconn;
  ZorpSatyrAuthenticator *self;
  PyObject *methods;
  gint port;
  
  if (!PyArg_ParseTuple(args, "OOi", &authconn, &methods, &port))
    return NULL;
  if (!z_py_zorp_authconn_check(authconn))
    {
      PyErr_SetString(PyExc_TypeError, "ZorpSatyrAuthenticator expected");
      return NULL;
    }
  self = PyObject_NEW(ZorpSatyrAuthenticator, &z_py_zorp_satyr_authenticator_type);

  self->authconn = authconn;
  Py_XINCREF(authconn);
  self->port = port;
  
  /* convert methods */
  self->methods[0] = "SKEY";
  self->methods[1] = "PASS";
  self->methods[2] = NULL;
  
  self->satyr = z_satyr_authenticator_new(authconn->authconn);
  if (!self->satyr)
    {
      PyErr_SetString(PyExc_IOError, "Error creating satyr instance");
      Py_XDECREF(self);
      return NULL;
    }
  return (PyObject *) self;
}

static void
z_py_zorp_satyr_authenticator_free(ZorpSatyrAuthenticator *self)
{
  if (self->satyr)
    {
      z_satyr_authenticator_unref(self->satyr);
      self->satyr = NULL;
    }
  Py_XDECREF(self->authconn);
  PyMem_DEL(self);
}

static PyObject *
z_py_zorp_satyr_authenticator_auth(ZorpSatyrAuthenticator *self, PyObject *args)
{
  gchar *desc;
  guint type;
  ZSatyrSubjID subj;
  ZorpSockAddr *client_addr;
  ZLabelSet *ls;
  
  if (!PyArg_ParseTuple(args, "sOi", &desc, &client_addr, &type))
    return NULL;
  
  subj.addr = client_addr->sa;
  subj.type = type;
  
  ls = z_satyr_authenticator_auth(self->satyr, &subj, self->port, desc, self->methods);
  if (ls)
    {
      return z_py_zorp_label_set_new(ls);
    }
  else
    {
      Py_XINCREF(Py_None);
      return Py_None;
    }
}

static PyMethodDef py_zorp_satyr_authenticator_methods[] =
{
  { "auth",        (PyCFunction) z_py_zorp_satyr_authenticator_auth, METH_VARARGS },
  { NULL,          NULL }   /* sentinel*/
};

static PyObject *
z_py_zorp_satyr_authenticator_getattr(ZorpSatyrAuthenticator *self, char *name)
{
  return Py_FindMethod(py_zorp_satyr_authenticator_methods, (PyObject *) self, name);
}

static int
z_py_zorp_satyr_authenticator_setattr(ZorpSatyrAuthenticator *self, char *name, PyObject *value)
{
  return -1;
}

PyTypeObject z_py_zorp_satyr_authenticator_type =
{
  PyObject_HEAD_INIT(&PyType_Type)
  0,                                        /*ob_size*/
  "ZorpSatyrAuthenticator",                 /*tp_name*/
  sizeof(ZorpSatyrAuthenticator),           /*tp_basicsize*/
  0,                                        /*tp_itemsize*/
  /* methods */
  (destructor)z_py_zorp_satyr_authenticator_free,      /*tp_dealloc*/
  (printfunc)0,                             /*tp_print*/
  (getattrfunc)z_py_zorp_satyr_authenticator_getattr,  /*tp_getattr*/
  (setattrfunc)z_py_zorp_satyr_authenticator_setattr,  /*tp_setattr*/
  (cmpfunc)0,                               /*tp_compare*/
  (reprfunc)0,                              /*tp_repr*/
  0,                                        /*tp_as_number*/
  0,                   		            /*tp_as_sequence*/
  0,                   		            /*tp_as_mapping*/
  (hashfunc)0,         		            /*tp_hash*/
  (ternaryfunc)0,      		            /*tp_call*/
  (reprfunc)0,         		            /*tp_str*/

  /* Space for future expansion */
  0L,0L,0L,0L,
  "ZorpSatyrAuthenticator class", /* Documentation string */
};

PyMethodDef z_py_zorp_satyr_authenticator_funcs[] =
{
  { "SatyrAuthenticator",  z_py_zorp_satyr_authenticator_new_instance, METH_VARARGS },
  { NULL,      NULL }   /* sentinel*/
};


void
z_py_zorp_satyr_authenticator_init(void)
{
  Py_InitModule("Zorp.Zorp", z_py_zorp_satyr_authenticator_funcs);
}
