/***************************************************************************
 *
 * Copyright (c) 2000 BalaBit IT Ltd, Budapest, Hungary
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Id: stream.c,v 1.15 2000/10/10 18:25:58 bazsi Exp $
 *
 * Author  : Bazsi
 * Auditor : kisza
 * Last version: 1.9
 * Notes:
 *   - comments
 *
 * Author  : Bazsi
 * Auditor : kisza
 * Last version: 1.9
 * Notes:
 *   - comments
 *
 ***************************************************************************/

#include <zorp/stream.h>
#include <zorp/log.h>

#include <string.h>
#include <errno.h>
#include <sys/socket.h>

void
z_data_dump(char *prefix, char *buf, guint len)
{
  guint i, j;
  gchar line[1024];
  
  i = 0;
  while (i < len)
    {
      char *end;
      
      end = line + g_snprintf(line, sizeof(line), "%s: ", prefix);
      
      for (j = 0; j < 16 && (i + j < len); j++)
        {
          g_snprintf(end, sizeof(line) - (end - line), "%02X ", (unsigned char) buf[i+j]);
          end += 3;
        }
      i += j;
      z_debug(10, "%s", line);
    }
}

typedef ZStream ZRealStream;

void 
z_stream_shutdown(ZStream *s, int i)
{
  int res, attempt = 1, saved_errno;
  
  z_debug(10, "%s: shutdown channel, mode=%d", s->name, i);
  
  errno = 0;
  do
    {
      res = shutdown(s->fd, i);
      saved_errno = errno;
      if (res == -1)
        {
          z_warning(0, "%s: shutdown failed, attempt=%d, %m", s->name, attempt++);
        }
    }
  while (saved_errno == EINTR);
}

static GIOError
z_stream_read_method(GIOChannel *channel,
		     gchar *buf,
		     guint count,
		     guint *bytes_read)
{
  ZRealStream *self = (ZRealStream *) channel;
  GIOError res;
  int attempt = 1, saved_errno;

  do
    {  
      res = g_io_channel_read(self->channel, buf, count, bytes_read);
      saved_errno = errno;
      if (res != G_IO_ERROR_NONE)
        {
          z_warning(5, "%s: read failed, attempt=%d: %m", self->name, attempt++);
        }
    }
  while (res == G_IO_ERROR_AGAIN && saved_errno == EINTR);

  z_debug(9, "%s: reading channel, fd=%d, count=%d", self->name, self->fd, *bytes_read);

  if (verbose_level >= 11) 
    {
      z_data_dump(self->name, buf, *bytes_read);
    }
  return res;
}

static GIOError
z_stream_write_method(GIOChannel *channel,
		      gchar *buf,
		      guint count,
		      guint *bytes_written)
{
  ZRealStream *self = (ZRealStream *) channel;
  GIOError res;
  int attempt = 1, saved_errno;

  /* FIXME: todo, datadump */
  
  z_debug(10, "%s: writing channel, fd=%d, count=%d", self->name, self->fd, count);
  if (verbose_level >= 11) 
    {
      z_data_dump(self->name, buf, *bytes_written);
    }
  do
    {
      res = g_io_channel_write(self->channel, buf, count, bytes_written);
      saved_errno = errno;
      if (res != G_IO_ERROR_NONE)
        {
          z_warning(5, "%s: write failed, attempt=%d: %m", self->name, attempt++);
        }
    }
  while (res == G_IO_ERROR_AGAIN && saved_errno == EINTR);
  return res;
}

static void
z_stream_close_method(GIOChannel *channel)
{
  ZRealStream *self = (ZRealStream *) channel;

  z_debug(10, "%s: closing channel, fd=%d", self->name, self->fd);
  g_io_channel_close(self->channel);
}

static void
z_stream_free(GIOChannel *channel)
{
  ZRealStream *self = (ZRealStream *) channel;

  g_io_channel_unref(self->channel);
  g_free(channel);
}

GIOFuncs z_stream_funcs =
{
  z_stream_read_method,
  z_stream_write_method,
  NULL,
  z_stream_close_method,
  NULL,
  z_stream_free
};

/*+

  Allocate and initialize a ZThread instance with the given fd and name.

  Parameters:
    fd        fd to wrap this stream around
    name      name to identify to stream in logs
  
  Returns:
    ZStream instance

  +*/
ZStream *
z_stream_new(gint fd, gchar *name)
{
  ZRealStream *self = g_new0(ZRealStream, 1);

  g_io_channel_init(&self->super);
  self->fd = fd;
  self->super.funcs = &z_stream_funcs;

  self->channel = g_io_channel_unix_new(fd);
  strncpy(self->name, name, sizeof(self->name) - 1);
  return (ZStream *) self;
}

