/***************************************************************************
 *
 * Copyright (c) 2000 BalaBit IT Ltd, Budapest, Hungary
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Id: thread.c,v 1.8 2000/10/10 18:25:58 bazsi Exp $
 *
 * Author  : Bazsi
 * Auditor : kisza
 * Last version: 1.6
 * Notes:
 *
 * Author  : Bazsi
 * Auditor : kisza
 * Last version: 1.6
 * Notes:
 *
 ***************************************************************************/

#include <zorp/thread.h>
#include <zorp/log.h>
#include <string.h>

GList *threads;

static void
z_thread_destroy(ZThread *self)
{
  threads = g_list_remove(threads, self);
  g_free(self);
}

/*+ 
  A function wrapped around the real thread function logging two
  events when the thread starts & stops 
  +*/
static void
z_thread_func(gpointer st)
{
  ZThread *self = (ZThread *) st;
  z_message(5, "%s: thread starting", self->name);
  (*self->func)(self->arg);
  z_message(5, "%s: thread exiting", self->name);
  z_thread_destroy(self);
}

/*+

  Allocate and initialize a Zorp thread identified by a name, and
  using the given thread function.

  Parameters:
    name      name to identify this thread with
    func      thread function
    arg       pointer to pass to thread function

  Returns:
    ZThread instance

  +*/
ZThread *
z_thread_new(gchar *name, GThreadFunc func, gpointer arg)
{
  ZThread *self = g_new0(ZThread, 1);
  static gint thread_id = 1;
  
  self->thread_id = thread_id++;
  self->func = func;
  self->arg = arg;
  strncpy(self->name, name, sizeof(self->name) - 1);
  self->thread = g_thread_create(z_thread_func, self, 0, 0, 0, G_THREAD_PRIORITY_NORMAL);
  threads = g_list_append(threads, self);
  return self;
}
