/***************************************************************************
 *
 * Copyright (c) 2000 BalaBit IT Ltd, Budapest, Hungary
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Id: http.c,v 1.19 2000/09/25 12:11:25 bazsi Exp $
 *
 * By: Kovacs Viktor Peter (vps__@hotmail.com | vps__@freemail.hu | ...)
 *
 ***************************************************************************/
   
#include "http.h"
#include "httpfltr.h"

//#define TRUE       (1) => not needed :-)
//#define FALSE      (0) -||-

//                (ok: 1,2,3 )
int http_connect(HttpProxy* s, HttpStr* addr);
int http_close_imminent(HttpProxy* s, int side);
int http_close_graceful(HttpProxy* s, int side);

//                (ok: 1,2,3,4,5,6 ) -> moved to http.h
//int http_recv(HttpProxy* s, int side);
//int http_send(HttpProxy* s, int side, char* buf, int len);
//int http_read(HttpProxy* s, int side, HttpStr* buf);
//int http_write(HttpProxy* s, int side, HttpStr* buf);
//int http_readline(HttpProxy* s, int side, HttpStr* buf, int maxlen);
//int http_writeline(HttpProxy* s, int side, HttpStr* buf);

//                (ok: 1,2,3 )
int  http_sendfile(HttpProxy* s, int fileid);
void http_logwrite(HttpProxy* s, int from, char* msg);
void http_log(HttpProxy* s, int eventclass, char* logstr);

//                (ok: 1,2,3,4,5,6 )
int http_parse_header(HttpStr* src, HttpStr* name, HttpStr* val);
int http_construct_header(HttpStr* dst, HttpStr* name, HttpStr* val);
int http_filter_hash_compare(gconstpointer a, gconstpointer b);
int http_filter_hash_bucket(gconstpointer a);
int http_filter_init(HttpProxy* s);
int http_filter(HttpProxy* s, int what, HttpStr* buf, HttpStr* url);

//----------------------------------------------------------------
int http_connect(HttpProxy* s, HttpStr* addr)
{ //should append :80 if nothing present
 HttpStr  tsd;
 HttpStr* ts=&tsd;
 int t=0, cport=80;
 
 if((s->transp_addr)==TRUE)
 {
  if(!z_proxy_connect_server_event(&s->super, "", 0)) return(HTTP_ERROR); 
  s->sstat=TRUE;
  http_str_sets(s->old_addr, "");
  http_str_sets(s->addr, "");
  return(0);
 }
 
 http_str_init(ts);
 while((addr->str)[t]!=0 && (addr->str)[t]!=':')
 {
  http_str_addc(ts, (addr->str)[t]);
  t++;
 }
 if((addr->str)[t]==':')
 {
  if(sscanf((addr->str)+t+1, "%u", &cport)!=1)
  {
   http_str_free(ts);
   return(HTTP_ERROR);
  }
  if((cport>65536) || (cport<0))
  {
   http_str_free(ts);
   return(HTTP_ERROR);
  }
 }
 if(!z_proxy_connect_server_event(&s->super, ts->str, cport))
 {
  http_str_free(ts);
  return(HTTP_ERROR);
 }
 s->sstat=TRUE;
 http_str_set(s->old_addr, s->addr);
 http_str_free(ts);
 return(0);
}

int http_close_imminent(HttpProxy* s, int side)
{
 if (side & HTTP_SIDE_CLIENT)
 {
  z_stream_close(s->super.endpoints[EP_CLIENT]);
  z_stream_unref(s->super.endpoints[EP_CLIENT]);
  s->super.endpoints[EP_CLIENT] = NULL;
  s->cstat=FALSE;
 }
 if (side & HTTP_SIDE_SERVER)
 {
  z_stream_close(s->super.endpoints[EP_SERVER]);
  z_stream_unref(s->super.endpoints[EP_SERVER]);
  s->super.endpoints[EP_SERVER] = NULL;
  s->sstat=FALSE;
 }
 return(0);
}

int http_close_graceful(HttpProxy* s, int side)
{
 /*int bytes_written=0;*/
 /*char buf[1]={0};*/
 
 if (side & HTTP_SIDE_CLIENT)
 {
  z_stream_shutdown(s->super.endpoints[EP_CLIENT], SHUT_RDWR);
 }
 if (side & HTTP_SIDE_SERVER)
 {
  z_stream_shutdown(s->super.endpoints[EP_SERVER], SHUT_RDWR);
 }
 return http_close_imminent(s, side);
#if 0
 switch(side)
 {
  case HTTP_SIDE_CLIENT:
   /* z_stream_write(s->super.endpoints[EP_CLIENT], buf, 0, &bytes_written); */
   z_stream_shutdown(s->super.endpoints[EP_CLIENT], SHUT_RDWR);
   z_stream_close(s->super.endpoints[EP_CLIENT]);
   s->cstat=FALSE;
   break;
  case HTTP_SIDE_SERVER:
   /* z_stream_write(s->super.endpoints[EP_SERVER], buf, 0, &bytes_written); */
   z_stream_shutdown(s->super.endpoints[EP_SERVER], SHUT_RDWR);
   z_stream_close(s->super.endpoints[EP_SERVER]);
   s->sstat=FALSE;
   break;
  case HTTP_SIDE_BOTH:
   /* z_stream_write(s->super.endpoints[EP_CLIENT], buf, 0, &bytes_written); */
   z_stream_shutdown(s->super.endpoints[EP_CLIENT], SHUT_RDWR);
   z_stream_close(s->super.endpoints[EP_CLIENT]);
   s->cstat=FALSE;
   /* z_stream_write(s->super.endpoints[EP_SERVER], buf, 0, &bytes_written); */
   z_stream_shutdown(s->super.endpoints[EP_SERVER], SHUT_RDWR);
   z_stream_close(s->super.endpoints[EP_SERVER]);
   s->sstat=FALSE;
   break;
 }
#endif
 return(0);
}

//----------------------------------------------------------------
int http_less(int a, int b)  // :-)
{
 if((a<0) || (b<0)) return(0);
 if(a > b) return(b);
 return(a); 
}

int http_recv(HttpProxy* s, int side)
{
 if(side==HTTP_SIDE_CLIENT) //client
 {
  s->bufcpos=0;
  if(z_stream_read(s->super.endpoints[EP_CLIENT], s->bufc, HTTP_BLOCKSIZE, &(s->bufclen))!=G_IO_ERROR_NONE) return(HTTP_ERROR);
  //z_debug(0, "IN: %d bytes", s->bufclen);
  return(s->bufclen);
 }else if(side==HTTP_SIDE_SERVER) //server
 {
  s->bufspos=0;
  if(z_stream_read(s->super.endpoints[EP_SERVER], s->bufs, HTTP_BLOCKSIZE, &(s->bufslen))!=G_IO_ERROR_NONE) return(HTTP_ERROR);
  //z_debug(0, "IN: %d bytes", s->bufslen);
  return(s->bufslen);
 }
 return(HTTP_ERROR);
}

int http_send(HttpProxy* s, int side, char* buf, int len)
{
 int bytes_written=0;
 //z_debug(0, "OUT: %d bytes", len);
 if(side==HTTP_SIDE_CLIENT)
 {
  if(z_stream_write(s->super.endpoints[EP_CLIENT], buf, len, &bytes_written)!=G_IO_ERROR_NONE) return(HTTP_ERROR);
  if(len!=bytes_written) return(HTTP_ERROR);
  return(bytes_written);
 }else if(side==HTTP_SIDE_SERVER)
 {
  if(z_stream_write(s->super.endpoints[EP_SERVER], buf, len, &bytes_written)!=G_IO_ERROR_NONE) return(HTTP_ERROR);
  if(len!=bytes_written) return(HTTP_ERROR);
  return(bytes_written);
 }
 return(HTTP_ERROR);
}

int http_read(HttpProxy* s, int side, HttpStr* buf)
{
 int t;

 if(side==HTTP_SIDE_CLIENT) //client
 {
  if((s->bufcpos)>=(s->bufclen)){if(http_recv(s, side)==HTTP_ERROR) return(HTTP_ERROR);}  //empty? => read
  if((t=http_less((s->bufclen)-(s->bufcpos), buf->len))==0) return(0);                    //t=less(remaining, buffer)
  memcpy(buf->str, (s->bufc)+(s->bufcpos), t);                                            //copy s->buf => buf (t)
  (s->bufcpos)+=t;                                                                        //pos+=t
  return(t);
 }else if(side==HTTP_SIDE_SERVER) //server
 {
  if((s->bufspos)>=(s->bufslen)){if(http_recv(s, side)==HTTP_ERROR) return(HTTP_ERROR);}
  if((t=http_less((s->bufslen)-(s->bufspos), buf->len))==0) return(0);
  memcpy(buf->str, (s->bufs)+(s->bufspos), t);
  (s->bufspos)+=t;
  return(t);   
 }
 return(HTTP_ERROR);
}

int http_write(HttpProxy* s, int side, HttpStr* buf)
{
 return(http_send(s, side, buf->str, buf->len));
}

int http_getch(HttpProxy* s, int side, int peekonly)
{
 int c;
 if(side==HTTP_SIDE_CLIENT) //client
 {
  if(s->bufcpos>=s->bufclen){if(http_recv(s, side)==HTTP_ERROR) return(HTTP_ERROR);}
  if(s->bufcpos==s->bufclen) return(HTTP_ERROR); //got eof
  c=(s->bufc)[s->bufcpos];
  if(peekonly==FALSE)(s->bufcpos)++;
  return(c);
 }else if(side==HTTP_SIDE_SERVER) //server
 {
  if(s->bufspos>=s->bufslen){if(http_recv(s, side)==HTTP_ERROR) return(HTTP_ERROR);}
  if(s->bufspos==s->bufslen) return(HTTP_ERROR); //got eof
  c=(s->bufs)[s->bufspos];
  if(peekonly==FALSE)(s->bufspos)++;
  return(c);
 }
 return(HTTP_ERROR);
}

int http_readline(HttpProxy* s, int side, HttpStr* buf, int maxlen)
{
 int t=0,c;
 
 http_str_sets(buf, "");
 while(t<maxlen)
 {
  if((c=http_getch(s, side, FALSE))==HTTP_ERROR) return(HTTP_ERROR);
  //if((c<32) && (c!='\n') && (c!='\r')) return(HTTP_ERROR); //filter
  if(c==0) return(HTTP_ERROR); //the header can NOT contain 0-s !!!
  switch(c)
  {
   case '\n':
    return(t); // ok
    break;
   case '\r': // ignore '\r'-s
    break;
   default:
    http_str_addc(buf, c); t++;
    break;
  }
 }
 return(HTTP_ERROR);
}

int http_writeline(HttpProxy* s, int side, HttpStr* buf)
{
 if((buf->len)>0) if(http_send(s, side, buf->str, buf->len)==HTTP_ERROR) return(HTTP_ERROR);
 if(http_send(s, side, "\r\n", 2)==HTTP_ERROR) return(HTTP_ERROR);
 return((buf->len)+2);
}

//----------------------------------------------------------------
int http_sendfile(HttpProxy* s, int fileid)
{
 HttpStr td;
 HttpStr* t=&td;
 HttpStr ntd;
 HttpStr* nt=&ntd;
 char* buf;
 int buflen=0;
 int f;
 struct stat st;
 
 http_str_init(t);
 http_str_init(nt);
 http_str_adds(nt, ZORP_DATADIR);
 http_str_adds(nt, "/http/");
 switch(fileid)
 {
  case HTTP_EFID_INTERNAL:
   http_str_sets(t, "HTTP/1.0 500 Internal Server Error\r\n");
   http_str_adds(nt, "efid_internal.htm");
   break;
  case  HTTP_EFID_SYNTAX:
   http_str_sets(t, "HTTP/1.0 400 Bad Request\r\n");
   http_str_adds(nt, "efid_syntax.htm");
   break;
  case  HTTP_EFID_CONNECT:
   http_str_sets(t, "HTTP/1.0 404 Not Found\r\n");
   http_str_adds(nt, "efid_connect.htm");
   break;
  case  HTTP_EFID_SERVER:
   http_str_sets(t, "HTTP/1.0 502 Bad Gateway\r\n");
   http_str_adds(nt, "efid_server.htm");
   break;
  case  HTTP_EFID_POLICY:
   http_str_sets(t, "HTTP/1.0 403 Forbidden\r\n");   
   http_str_adds(nt, "efid_policy.htm");
   break;
  default:
   http_str_sets(t, "HTTP/1.0 200 OK\r\n");
   http_str_adds(nt, "efid_default.htm");
   break;
 }
 if(s->ver!=0x0009)
 {
  http_str_adds(t, "Content-Type: text/html\r\n");
 }
 buflen=http_writeline(s, HTTP_SIDE_CLIENT, t);
 http_str_free(t);
 if(buflen==HTTP_ERROR)
 {
  http_str_free(nt);
  http_close_imminent(s, HTTP_SIDE_CLIENT);
  return(HTTP_ERROR);
 }

 if((f=open(nt->str, O_RDONLY))==-1)
 {
  http_str_adds(nt, "l");
  if((f=open(nt->str, O_RDONLY))==-1)
  {
   http_str_free(nt); //no error file => use this
   buf = "<html><head><title>PROXY-ERROR</title></head><body><b>Internal Proxy Error!</b><br></body></html>";
   buflen = strlen(buf);
   http_send(s, HTTP_SIDE_CLIENT, buf, buflen);
   return(HTTP_ERROR);
  }
 }
 fstat(f, &st); buflen=st.st_size;
 http_str_allocfill(nt, buflen, 0);
 read(f, nt->str, buflen);
 close(f);
 
 if(http_send(s, HTTP_SIDE_CLIENT, nt->str, buflen)==HTTP_ERROR)
 {
  http_str_free(nt);
  http_close_imminent(s, HTTP_SIDE_CLIENT);
  return(HTTP_ERROR);
 }

 http_str_free(nt); 
 http_close_graceful(s, HTTP_SIDE_CLIENT);
 return(0);
}

char b2h[]="0123456789ABCDEF";

void http_logwrite(HttpProxy* s, int from, char* msg)
{
// error level mapping: (direct)
 char buf[HTTP_LOGLINESIZE+1]=""; //+1:terminator
 int errl=0, t=0, tc=0;
 switch(from)
 {
  case HTTP_LOG_ERR0:
   errl=0;
   break;
  case HTTP_LOG_ERR1:
   errl=1;
   break;
  case HTTP_LOG_ERR2:
   errl=2;
   break;
  case HTTP_LOG_ERR3:
   errl=7;
   break;
  case HTTP_LOG_EVN0:
   errl=0;
   break;
  case HTTP_LOG_EVN1:
   errl=5;
   break;
  case HTTP_LOG_EVN2:
   errl=7;
   break;
  case HTTP_LOG_EVN3:
   errl=9;
   break;
 }
 
 //nonprintable filtering
 while(msg[t]!=0 && (tc<HTTP_LOGLINESIZE - 4))
 {
  if(msg[t]<' ')
  {
   buf[tc++]='\\';
   buf[tc++]='x';
   buf[tc++]=b2h[(msg[t] >>4) & 15];
   buf[tc++]=b2h[msg[t] & 15];
   t++;
  }else
  {
   buf[tc++]=msg[t++]; 
  } 
 }
 buf[tc]=0;
 
 if(from & (HTTP_LOG_ERR0 | HTTP_LOG_ERR1 | HTTP_LOG_ERR2 | HTTP_LOG_ERR3))
  z_warning(errl, "%s: %s", s->super.session_id, buf); 
 else
  z_message(errl, "%s: %s", s->super.session_id, buf);
}

void http_log(HttpProxy* s, int eventclass, char* logstr)
{
 char buf[HTTP_LOGLINESIZE+1]=""; //+1:terminator
 char moc=0;
 if(logstr==NULL) logstr=&moc;

 switch(eventclass)
 {
  case HTTP_ERR_INTERNAL:     //0
   if(((s->logmode) & HTTP_LOG_ERR0) == 0) break;
   g_snprintf(buf, HTTP_LOGLINESIZE, "HTTP ERROR : INTERNAL");
   http_logwrite(s, HTTP_LOG_ERR0, buf);
   break;
  case HTTP_ERR_FATAL:        //0
   if(((s->logmode) & HTTP_LOG_ERR0) == 0) break;
   g_snprintf(buf, HTTP_LOGLINESIZE, "HTTP ERROR : FATAL");
   http_logwrite(s, HTTP_LOG_ERR0, buf);
   break;
  case HTTP_ERR_UNKNOWN:      //0
   if(((s->logmode) & HTTP_LOG_ERR0) == 0) break;
   g_snprintf(buf, HTTP_LOGLINESIZE, "HTTP ERROR : UNKNOWN");
   http_logwrite(s, HTTP_LOG_ERR0, buf);
   break;
  case HTTP_ERR_CLIENT_FAIL:  //3
   if(((s->logmode) & HTTP_LOG_ERR3) == 0) break;
   g_snprintf(buf, HTTP_LOGLINESIZE, "HTTP ERROR : CLIENT FAILURE");
   http_logwrite(s, HTTP_LOG_ERR3, buf);
   break;
  case HTTP_ERR_SERVER_FAIL:  //2
   if(((s->logmode) & HTTP_LOG_ERR2) == 2) break;
   g_snprintf(buf, HTTP_LOGLINESIZE, "HTTP ERROR : SERVER FAILURE (%s)", logstr);
   http_logwrite(s, HTTP_LOG_ERR2, buf);
   break;
  case HTTP_ERR_CONNECT:      //2
   if(((s->logmode) & HTTP_LOG_ERR2) == 0) break;
   g_snprintf(buf, HTTP_LOGLINESIZE, "HTTP ERROR : SERVER CONNECT (%s)", s->addr->str);
   http_logwrite(s, HTTP_LOG_ERR2, buf);
   break;
  case HTTP_ERR_CLIENT_SYNTAX://2
   if(((s->logmode) & HTTP_LOG_ERR0) == 0) break;
   g_snprintf(buf, HTTP_LOGLINESIZE, "HTTP ERROR : SYNTAX ERROR IN REQUEST (%s)", logstr);
   http_logwrite(s, HTTP_LOG_ERR2, buf);
   break;
  case HTTP_ERR_SERVER_SYNTAX://2
   if(((s->logmode) & HTTP_LOG_ERR2) == 0) break;
   g_snprintf(buf, HTTP_LOGLINESIZE, "HTTP ERROR : SYNTAX ERROR IN RESPONSE");
   http_logwrite(s, HTTP_LOG_ERR2, buf);
   break;
  case HTTP_ERR_POLICY:       //0
   if(((s->logmode) & HTTP_LOG_ERR0) == 0) break;
   g_snprintf(buf, HTTP_LOGLINESIZE, "HTTP ERROR : POLICY CALL FAILED");
   http_logwrite(s, HTTP_LOG_ERR0, buf);
   break;

  case HTTP_EVENT_INIT:       //0
   if(((s->logmode) & HTTP_LOG_EVN0) == 0) break;
   g_snprintf(buf, HTTP_LOGLINESIZE, "http event : proxy init");
   http_logwrite(s, HTTP_LOG_EVN0, buf);
   break;
  case HTTP_EVENT_SHUTDOWN:   //0
   if(((s->logmode) & HTTP_LOG_EVN0) == 0) break;
   g_snprintf(buf, HTTP_LOGLINESIZE, "http event : proxy shutdown");
   http_logwrite(s, HTTP_LOG_EVN0, buf);
   break;
  case HTTP_EVENT_ACCEPT:     //1
   if(((s->logmode) & HTTP_LOG_EVN1) == 0) break;
   g_snprintf(buf, HTTP_LOGLINESIZE, "http event : client connection accepted");
   http_logwrite(s, HTTP_LOG_EVN1, buf);
   break;
  case HTTP_EVENT_CONNECT:    //1
   if(((s->logmode) & HTTP_LOG_EVN1) == 0) break;
   g_snprintf(buf, HTTP_LOGLINESIZE, "http event : connection to server (%s)", s->addr->str);
   http_logwrite(s, HTTP_LOG_EVN1, buf);
   break;
  case HTTP_EVENT_CLOSE_CLIENT_IMMINENT: //1
   if(((s->logmode) & HTTP_LOG_EVN1) == 0) break;
   g_snprintf(buf, HTTP_LOGLINESIZE, "http event : imminent close of client connection");
   http_logwrite(s, HTTP_LOG_EVN1, buf);
   break;
  case HTTP_EVENT_CLOSE_CLIENT_GRACEFUL: //1
   if(((s->logmode) & HTTP_LOG_EVN1) == 0) break;
   g_snprintf(buf, HTTP_LOGLINESIZE, "http event : graceful close of client connection");
   http_logwrite(s, HTTP_LOG_EVN1, buf);
   break;
  case HTTP_EVENT_CLOSE_SERVER_IMMINENT: //1
   if(((s->logmode) & HTTP_LOG_EVN1) == 0) break;
   g_snprintf(buf, HTTP_LOGLINESIZE, "http event : imminent close of server connection");
   http_logwrite(s, HTTP_LOG_EVN1, buf);
   break;
  case HTTP_EVENT_CLOSE_SERVER_GRACEFUL: //1
   if(((s->logmode) & HTTP_LOG_EVN1) == 0) break;
   g_snprintf(buf, HTTP_LOGLINESIZE, "http event : graceful close of server connection");
   http_logwrite(s, HTTP_LOG_EVN1, buf);
   break;
  case HTTP_EVENT_FILTER:     //2
   if(((s->logmode) & HTTP_LOG_EVN2) == 0) break; //unused
   g_snprintf(buf, HTTP_LOGLINESIZE, "http event : filtering");
   http_logwrite(s, HTTP_LOG_EVN2, buf);
   break;
  case HTTP_EVENT_FILTER_IN:  //3
   if(((s->logmode) & HTTP_LOG_EVN3) == 0) break;
   g_snprintf(buf, HTTP_LOGLINESIZE, "http event : filter input: %s", logstr);
   http_logwrite(s, HTTP_LOG_EVN3, buf);
   break;
  case HTTP_EVENT_FILTER_OUT:  //3
   if(((s->logmode) & HTTP_LOG_EVN3) == 0) break;
   g_snprintf(buf, HTTP_LOGLINESIZE, "http event : filter output: %s", logstr);
   http_logwrite(s, HTTP_LOG_EVN3, buf);
   break;
  case HTTP_EVENT_POLICY_CALL://3
   if(((s->logmode) & HTTP_LOG_EVN3) == 0) break;
   g_snprintf(buf, HTTP_LOGLINESIZE, "http event : policy call (%s)", logstr);
   http_logwrite(s, HTTP_LOG_EVN3, buf);
   break;
  case HTTP_EVENT_REQUEST:    //2
   if(((s->logmode) & HTTP_LOG_EVN2) == 0) break;
   g_snprintf(buf, HTTP_LOGLINESIZE, "http event : reading request");
   http_logwrite(s, HTTP_LOG_EVN2, buf);
   break;
  case HTTP_EVENT_RESPONSE:   //2
   if(((s->logmode) & HTTP_LOG_EVN2) == 0) break;
   g_snprintf(buf, HTTP_LOGLINESIZE, "http event : reading response");
   http_logwrite(s, HTTP_LOG_EVN2, buf);
   break;
  case HTTP_EVENT_COMPLETE:   //1
   if(((s->logmode) & HTTP_LOG_EVN1) == 0) break;
   g_snprintf(buf, HTTP_LOGLINESIZE, "http event : transfer complete");
   http_logwrite(s, HTTP_LOG_EVN1, buf);
   break;
  default:
   break;
 }
 return;
}
//----------------------------------------------------------------
int http_parse_header(HttpStr* src, HttpStr* name, HttpStr* val)
{  //format: "name: val"
 int t=0;

 http_str_sets(name, ""); http_str_sets(val, "");
 while((src->str)[t]!=0 && (src->str)[t]==' ') t++;                   //skip space
 while((src->str)[t]!=0 && (src->str)[t]!=' ' && (src->str)[t]!=':')  //copy to name till 0 | ' ' | ':'
  http_str_addc(name, (src->str)[t++]);
 if((src->str)[t]!=':')                                               //skip till ':'
  while((src->str)[t]!=0 && (src->str)[t]!=':') t++;
 if((src->str)[t]==':') t++;                                          //if ':' skip
 while((src->str)[t]!=0 && (src->str)[t]==' ') t++;                   //skip space
 while((src->str)[t]!=0)                                              //copy to end 
  http_str_addc(val, (src->str)[t++]);
 
 return(TRUE);
}

int http_construct_header(HttpStr* dst, HttpStr* name, HttpStr* val)
{  //format: "name: val"
 http_str_sets(dst, "");   http_str_add(dst, name);
 http_str_adds(dst, ": "); http_str_add(dst, val);
 return(TRUE);
}

int http_filter_hash_compare(gconstpointer a, gconstpointer b)
{ //case insensitive compare
 if(strcasecmp((char*)a, (char*)b)==0) return(1);
 return(0);
}

int http_filter_hash_bucket(gconstpointer a)
{
 int sum=0;
 char* s=(char*)a;
 while(*s!=0)
 {
  sum+=toupper(*s); //case insensitive
  s++;
 }
 return(sum % 16);  //16 buckets
}

int http_filter_init(HttpProxy* s)
{
 ZPolicyObj* res=NULL;

 //add default commands
 if(g_hash_table_lookup(s->req_cmd, "GET")==NULL)
 {
   res=z_policy_var_build("i", HTTP_FILTER_PASS);
   g_hash_table_insert(s->req_cmd, "GET", res);
 }
 if(g_hash_table_lookup(s->req_cmd, "HEAD")==NULL)
 {
   res=z_policy_var_build("i", HTTP_FILTER_PASS);
   g_hash_table_insert(s->req_cmd, "HEAD", res);
 } 
 if(g_hash_table_lookup(s->req_cmd, "POST")==NULL)
 {
   res=z_policy_var_build("i", HTTP_FILTER_PASS);
   g_hash_table_insert(s->req_cmd, "POST", res);
 } 

 if((s->transp) == FALSE) //add non transparent translation data
 {
  if((s->conv2proxy) == FALSE) //false: proxy => client ; true: client => proxy
  {
   if(g_hash_table_lookup(s->req_head, "Proxy-Connection")==NULL)
   {
    res=z_policy_var_build("(iss)", HTTP_FILTER_CHANGE_NAME, "Connection", "");
    g_hash_table_insert(s->req_head, "Proxy-Connection", res);
   }
   if(g_hash_table_lookup(s->resp_head, "Connection")==NULL)
   {
    res=z_policy_var_build("(iss)", HTTP_FILTER_CHANGE_NAME, "Proxy-Connection", "");
    g_hash_table_insert(s->resp_head, "Connection", res);
   }
  }else     
  {
   if(g_hash_table_lookup(s->resp_head, "Proxy-Connection")==NULL)
   {
    res=z_policy_var_build("(iss)", HTTP_FILTER_CHANGE_NAME, "Connection", "");
    g_hash_table_insert(s->resp_head, "Proxy-Connection", res);
   }
   if(g_hash_table_lookup(s->req_head, "Connection")==NULL)
   {
    res=z_policy_var_build("(iss)", HTTP_FILTER_CHANGE_NAME, "Proxy-Connection", "");
    g_hash_table_insert(s->req_head, "Connection", res);
   }
  }  
 }
 return(TRUE);
}

int http_filter(HttpProxy* s, int what, HttpStr* buf, HttpStr* url)
{ //return value in buf
 ZPolicyObj* res;
 HttpStr  t1d;
 HttpStr* t1=&t1d;
 HttpStr  t2d;
 HttpStr* t2=&t2d; 
 char*    cp1;
 char*    cp2;
 int      flg;

 http_str_init(t1); http_str_init(t2);
 switch(what)
 {
  case HTTP_FILTER_WHAT_REQ_CMD: //request command
    res=g_hash_table_lookup(s->req_cmd, buf->str);
    if(res==NULL)
    {
     if(s->req_cmd_def == HTTP_FILTER_PASS) //not found: use default
     {
      http_str_free(t1); http_str_free(t2);     
      return(TRUE);
     }
     http_str_free(t1); http_str_free(t2);   
     return(FALSE);
    }
    if(z_policy_var_parse(res, "i", &flg) == 0) 
    {
     http_str_free(t1); http_str_free(t2);
     return(FALSE);      
    }
    switch(flg)
    {
     case HTTP_FILTER_PASS: //pass through
       break; 
     case HTTP_FILTER_POLICY: //call policy
       res = z_policy_call (s->super.handler, "filterReqCmd", z_policy_var_build ("(ss)", buf->str, url->str));
       if(z_policy_var_parse(res, "s", &cp1) == 0)
       {
        http_str_free(t1); http_str_free(t2);
        z_policy_var_unref (res);
        return(FALSE);
       }
       http_str_sets(buf, cp1);           
       z_policy_var_unref (res); 
       break;
     case HTTP_FILTER_DROP: //drop request
     default:
       http_str_free(t1); http_str_free(t2);
       return(FALSE);
       break;                 
    }
    http_str_free(t1); http_str_free(t2);        
    return(TRUE);
    break;
  case HTTP_FILTER_WHAT_REQ_HEAD:
    http_parse_header(buf, t1, t2);
    res=g_hash_table_lookup(s->req_head, t1->str);
    if(res==NULL)
    {
     if(s->req_head_def == HTTP_FILTER_PASS)
     {
      http_construct_header(buf, t1, t2);
      http_str_free(t1); http_str_free(t2);     
      return(TRUE);
     }
     http_str_free(t1); http_str_free(t2);   
     return(FALSE);
    }
    if(z_policy_var_parse(res, "(iss)", &flg, &cp1, &cp2) == 0) 
    {
     http_str_free(t1); http_str_free(t2);
     return(FALSE);      
    }
    switch(flg)
    {
     case HTTP_FILTER_PASS:
       http_construct_header(buf, t1, t2);
       break; 
     case HTTP_FILTER_CHANGE_BOTH:
       http_str_sets(t1, cp1); http_str_sets(t2, cp2);
       http_construct_header(buf, t1, t2);
       break;
     case HTTP_FILTER_CHANGE_NAME:
       http_str_sets(t1, cp1);
       http_construct_header(buf, t1, t2);
       break;
     case HTTP_FILTER_CHANGE_VALUE:
       http_str_sets(t2, cp2);
       http_construct_header(buf, t1, t2);
       break;
     case HTTP_FILTER_POLICY:
       res = z_policy_call (s->super.handler, "filterReqHead", z_policy_var_build ("(ss)", t1->str, t2->str));
       if(z_policy_var_parse(res, "s", &cp1) == 0)
       {
        http_str_free(t1); http_str_free(t2);
        z_policy_var_unref (res);
        return(FALSE);
       }
       http_str_sets(buf, cp1);           
       z_policy_var_unref (res); 
       break;
     case HTTP_FILTER_DROP:
     default:
       http_str_free(t1); http_str_free(t2);
       return(FALSE);
       break;                 
    }
    http_str_free(t1); http_str_free(t2);        
    return(TRUE);
    break;
  case HTTP_FILTER_WHAT_RESP_CMD:
    res=g_hash_table_lookup(s->resp_cmd, buf->str);
    if(res==NULL)
    {
     if(s->resp_cmd_def == HTTP_FILTER_PASS)
     {
      http_str_free(t1); http_str_free(t2);     
      return(TRUE);
     }
     http_str_free(t1); http_str_free(t2);   
     return(FALSE);
    }
    if(z_policy_var_parse(res, "i", &flg) == 0) 
    {
     http_str_free(t1); http_str_free(t2);
     return(FALSE);      
    }
    switch(flg)
    {
     case HTTP_FILTER_PASS:
       break; 
     case HTTP_FILTER_POLICY:
       res = z_policy_call (s->super.handler, "filterRespCmd", z_policy_var_build ("(s)", buf->str));
       if(z_policy_var_parse(res, "s", &cp1) == 0)
       {
        http_str_free(t1); http_str_free(t2);
        z_policy_var_unref (res);
        return(FALSE);
       }
       http_str_sets(buf, cp1);           
       z_policy_var_unref (res); 
       break;
     case HTTP_FILTER_DROP:
     default:
       http_str_free(t1); http_str_free(t2);
       return(FALSE);
       break;                 
    }
    http_str_free(t1); http_str_free(t2);        
    return(TRUE);
    break;
  case HTTP_FILTER_WHAT_RESP_HEAD:
    http_parse_header(buf, t1, t2);
    res=g_hash_table_lookup(s->resp_head, t1->str);
    if(res==NULL)
    {
     if(s->resp_head_def == HTTP_FILTER_PASS)
     {
      http_construct_header(buf, t1, t2);
      http_str_free(t1); http_str_free(t2);     
      return(TRUE);
     }
     http_str_free(t1); http_str_free(t2);   
     return(FALSE);
    }
    if(z_policy_var_parse(res, "(iss)", &flg, &cp1, &cp2) == 0) 
    {
     http_str_free(t1); http_str_free(t2);
     return(FALSE);      
    }
    switch(flg)
    {
     case HTTP_FILTER_PASS:
       http_construct_header(buf, t1, t2);
       break; 
     case HTTP_FILTER_CHANGE_BOTH:
       http_str_sets(t1, cp1); http_str_sets(t2, cp2);
       http_construct_header(buf, t1, t2);
       break;
     case HTTP_FILTER_CHANGE_NAME:
       http_str_sets(t1, cp1);
       http_construct_header(buf, t1, t2);
       break;
     case HTTP_FILTER_CHANGE_VALUE:
       http_str_sets(t2, cp2);
       http_construct_header(buf, t1, t2);
       break;
     case HTTP_FILTER_POLICY:
       res = z_policy_call (s->super.handler, "filterRespHead", z_policy_var_build ("(ss)", t1->str, t2->str));       
       if(z_policy_var_parse(res, "s", &cp1) == 0)
       {
        http_str_free(t1); http_str_free(t2);
        z_policy_var_unref (res);
        return(FALSE);
       }
       http_str_sets(buf, cp1);           
       z_policy_var_unref (res); 
       break;
     case HTTP_FILTER_DROP:
     default:
       http_str_free(t1); http_str_free(t2);
       return(FALSE);
       break;                 
    }
    http_str_free(t1); http_str_free(t2);        
    return(TRUE);
    break;
  default:
    http_str_free(t1); http_str_free(t2);
    return(FALSE);   
 }
 http_str_free(t1); http_str_free(t2);
 return(FALSE);
}
 
//----------------------------------------------------------------
int http_main(HttpProxy* s, HttpStr* tbuf, HttpStr* tnewbuf, HttpStr* t1, HttpStr* t2, HttpStr* t3, 
                                                             HttpStr* t4, HttpStr* t5, HttpStr* t6)
{
 int x=0,y=0;
 char* tbp=NULL;
 http_log(s, HTTP_EVENT_ACCEPT, NULL);
// 1. //config is gone... :-)

//loop for keep-alive
while(1==1)
{
// 2.
 http_log(s, HTTP_EVENT_REQUEST, NULL);
 http_str_sets(tbuf, ""); x=0; y=FALSE;
 while(x<(s->maxreqline))
 {
  if(http_readline(s, HTTP_SIDE_CLIENT, t1, s->maxlinelen)==HTTP_ERROR)
  {
   http_close_imminent(s, HTTP_SIDE_CLIENT);
   http_log(s, HTTP_ERR_CLIENT_FAIL, NULL);
   return(HTTP_ERR_CLIENT_FAIL);
  }
  if((t1->len)!=0) y=TRUE; //skip first empty lines
  if(((t1->len)==0) && (y==TRUE)) break;
  if(y==TRUE)
  {
   http_str_add(tbuf, t1);  
   http_str_adds(tbuf, "\n");
  }
  x++;
 }
 if(x==(s->maxreqline) || (tbuf->len)==0)
 {
  http_close_imminent(s, HTTP_SIDE_CLIENT);
  http_log(s, HTTP_ERR_CLIENT_FAIL, NULL);
  return(HTTP_ERR_CLIENT_FAIL);
 }
 http_str_copyn(t1, tbuf, http_str_contains_c(tbuf, '\n'));

 if(http_splitreq(t1, t2, t3, t4, t5, t6)==FALSE) //t1:src 2:cmd 3:addr 4:url 5:ver 6:user+pass
 {
  http_log(s, HTTP_ERR_CLIENT_SYNTAX, t1->str);
  if(http_sendfile(s, HTTP_EFID_SYNTAX)==HTTP_ERROR)
  {
   http_log(s, HTTP_ERR_CLIENT_FAIL, NULL);
   return(HTTP_ERR_CLIENT_FAIL);
  }
  return(HTTP_ERR_CLIENT_SYNTAX);
 }
 if(http_check_url(t4)==FALSE)
 {
  http_log(s, HTTP_ERR_CLIENT_SYNTAX, t1->str);
  if(http_sendfile(s, HTTP_EFID_SYNTAX)==HTTP_ERROR)
  {
   http_log(s, HTTP_ERR_CLIENT_FAIL, NULL);
   return(HTTP_ERR_CLIENT_FAIL);
  }
  return(HTTP_ERR_CLIENT_SYNTAX);              
 }
 //check cmd & url; '/', '*' addition + CONNECT breakout
 s->reqflg=http_get_req_param(t2->str); //get protocol dependent flags
 if(((t4->len)==0) && (((s->reqflg) & HTTP_REQ_FLG_ASTERIX) != 0))
 {
  http_str_sets(t4, "*");
 }else if((t4->len)==0)
 {
  http_str_sets(t4, "/");
 }

 //if connect => call the plug
 if(((s->reqflg) & HTTP_REQ_FLG_CONNECT) != 0)
 {
  z_policy_lock(s->super.thread);
  if(http_filter(s, HTTP_FILTER_WHAT_REQ_CMD, t2, t4)==TRUE)
  {
   z_policy_unlock(s->super.thread);
   if((s->sstat) == TRUE)
   {
    http_close_imminent(s, HTTP_SIDE_SERVER);
    s->sstat=FALSE;
   }
   http_log(s, HTTP_EVENT_CONNECT, s->addr->str);
   if(http_connect(s, t4)==HTTP_ERROR)
   {
    http_log(s, HTTP_ERR_CONNECT, NULL);
    if(http_sendfile(s, HTTP_EFID_CONNECT)==HTTP_ERROR)
    {
     http_log(s, HTTP_ERR_CLIENT_FAIL, NULL);
     return(HTTP_ERR_CLIENT_FAIL);
    }
    return(HTTP_ERR_CONNECT);
   }
  
   //my_zorp_replace_proxy(...) :-))))))
   
   http_log(s, HTTP_EVENT_COMPLETE, NULL);
   return(HTTP_EVENT_COMPLETE);
  }else
  {
   z_policy_unlock(s->super.thread);
   http_log(s, HTTP_ERR_POLICY, NULL);
   if(http_sendfile(s, HTTP_EFID_POLICY) == HTTP_ERROR)
   {
    http_log(s, HTTP_ERR_CLIENT_FAIL, NULL);
    return(HTTP_ERR_CLIENT_FAIL);
   }  
   return(HTTP_ERR_POLICY);  
  }
 }

 if(((t2->len)==0) || ((s->reqflg)==-1)) //bad command
 {
  http_log(s, HTTP_ERR_CLIENT_SYNTAX, NULL);
  if(http_sendfile(s, HTTP_EFID_SYNTAX)==HTTP_ERROR)
  {
   http_log(s, HTTP_ERR_CLIENT_FAIL, NULL);
   return(HTTP_ERR_CLIENT_FAIL);
  }
  return(HTTP_ERR_CLIENT_SYNTAX);
 }
 //check protocol version
 if(http_str_cmpis(t5, "HTTP/1.0")==TRUE)
 {
  s->mode=HTTP_MODE_CLOSE;
  s->ver=0x100;
 }else if(http_str_cmpis(t5, "HTTP/1.1")==TRUE)
 {
  s->mode=HTTP_MODE_KEEPALIVE;
  s->ver=0x101;
 }else if(http_str_cmpis(t5, "")==TRUE)
 {
  s->mode=HTTP_MODE_CLOSE;
  s->ver=0x009;
 }else{
  http_log(s, HTTP_ERR_CLIENT_SYNTAX, NULL);
  if(http_sendfile(s, HTTP_EFID_SYNTAX) == HTTP_ERROR)
  {
   http_log(s, HTTP_ERR_CLIENT_FAIL, NULL);
   return(HTTP_ERR_CLIENT_FAIL);
  }
  return(HTTP_ERR_CLIENT_SYNTAX);
 }
 
 //check protocol override
 if(http_lookup(tbuf, "Connection", t1, TRUE)==TRUE)
 {
  if(http_str_cmpis(t1, "Keep-Alive")==TRUE) s->mode=HTTP_MODE_KEEPALIVE;
  if(http_str_cmpis(t1, "Close")==TRUE)      s->mode=HTTP_MODE_CLOSE;
 }
 if(http_lookup(tbuf, "Proxy-Connection", t1, TRUE)==TRUE)
 {
  if(http_str_cmpis(t1, "Keep-Alive")==TRUE) s->mode=HTTP_MODE_KEEPALIVE;
  if(http_str_cmpis(t1, "Close")==TRUE) s->mode=HTTP_MODE_CLOSE;
 }
 //check server address
 if((t3->len)!=0)
 {
  http_str_set(s->addr, t3); //from cmd
 }else if(http_lookup(tbuf, "Host", t1, TRUE)==TRUE)
 {
  http_str_set(s->addr, t1); //from "Host"
 }else if(s->transp==TRUE)
 {
  http_str_sets(s->addr, ""); //not needed
 }else
 {
  http_log(s, HTTP_ERR_CLIENT_SYNTAX, NULL);
  if(http_sendfile(s, HTTP_EFID_SYNTAX)==HTTP_ERROR)
  {
   http_log(s, HTTP_ERR_CLIENT_FAIL, NULL);
   return(HTTP_ERR_CLIENT_FAIL);
  }
  return(HTTP_ERR_CLIENT_SYNTAX);
 }
 /* 
 //policy event for filter setup
 z_policy_lock(s->super.thread);
 x=z_policy_event(s->super.handler, "filter_setup", z_policy_var_build ("(ss)", "rc", s->addr->str));
 z_policy_unlock(s->super.thread);
 */ 

 http_filter_init(s); //init filter with protocol dependent datas

// 3. //filter
 http_log(s, HTTP_EVENT_FILTER_IN, tbuf->str);
 http_str_sets(tnewbuf, "");
 z_policy_lock(s->super.thread);
 if((s->transp) == FALSE)
 { //reformat request
  if(http_filter(s, HTTP_FILTER_WHAT_REQ_CMD, t2, t4)==FALSE)
  {
   z_policy_unlock(s->super.thread);
   http_log(s, HTTP_ERR_POLICY, NULL);
   if(http_sendfile(s, HTTP_EFID_POLICY) == HTTP_ERROR)
   {
    http_log(s, HTTP_ERR_CLIENT_FAIL, NULL);
    return(HTTP_ERR_CLIENT_FAIL);
   }  
   return(HTTP_ERR_POLICY);
  }
  z_policy_unlock(s->super.thread);
  http_str_add(tnewbuf, t2); http_str_adds(tnewbuf, " ");
  if((s->conv2proxy) == TRUE)
  {
   http_str_adds(tnewbuf, "http://");
   http_str_add(tnewbuf, s->addr); 
  }
  http_str_add(tnewbuf, t4); http_str_adds(tnewbuf, " ");
  http_str_add(tnewbuf, t5);
  http_str_adds(tnewbuf, "\r\n");
 }else
 { //keep original request format
  if(http_filter(s, HTTP_FILTER_WHAT_REQ_CMD, t2, t4)==FALSE)
  {
   z_policy_unlock(s->super.thread);
   http_log(s, HTTP_ERR_POLICY, NULL);
   if(http_sendfile(s, HTTP_EFID_POLICY) == HTTP_ERROR)
   {
    http_log(s, HTTP_ERR_CLIENT_FAIL, NULL);
    return(HTTP_ERR_CLIENT_FAIL);
   }  
   return(HTTP_ERR_POLICY);
  }
  z_policy_unlock(s->super.thread);
  x=0;
  while((tbuf->str)[x]!=0 && (tbuf->str)[x]!='\n')
  {
   http_str_addc(tnewbuf, (tbuf->str)[x]);
   x++;
  }
  http_str_adds(tnewbuf, "\r\n");
 }
 //z_policy_unlock(s->super.thread);
 
 x=http_str_contains_c(tbuf, '\n'); x++; //skip req
 http_str_sets(t1, "");
 while((tbuf->str)[x]!=0)
 {
  if((tbuf->str)[x]=='\n')
  {
   z_policy_lock(s->super.thread);
   if(http_filter(s, HTTP_FILTER_WHAT_REQ_HEAD, t1, NULL)==TRUE)
   {
    http_str_adds(t1, "\r\n");
    http_str_add(tnewbuf, t1);
    http_str_sets(t1, "");
   }
   z_policy_unlock(s->super.thread);
  }
  else
  {
   http_str_addc(t1, (tbuf->str)[x]);
  }
  x++;
 }
 
// 4.
 if((s->sstat==TRUE)) // && (http_str_cmpi(s->addr, s->old_addr)==FALSE)) //close if changed
 { //workaround (server close bug /former keep-alive bug/)
  http_close_imminent(s, HTTP_SIDE_SERVER);
  s->sstat=FALSE;
 }
 if(s->sstat==FALSE) //connect if not connected
 {
  http_log(s, HTTP_EVENT_CONNECT, s->addr->str);
  if(http_connect(s, s->addr)==HTTP_ERROR)
  {
   http_log(s, HTTP_ERR_CONNECT, NULL);
   if(http_sendfile(s, HTTP_EFID_CONNECT)==HTTP_ERROR)
   {
    http_log(s, HTTP_ERR_CLIENT_FAIL, NULL);
    return(HTTP_ERR_CLIENT_FAIL);
   }
   return(HTTP_ERR_CONNECT);
  }
  z_policy_lock(s->super.thread);
  x=z_policy_event(s->super.handler, "command", z_policy_var_build ("(ss)", "rc", t2->str));
  z_policy_unlock(s->super.thread);
  if(x == Z_ABORT)
  {
   http_log(s, HTTP_ERR_POLICY, NULL);
   http_close_imminent(s, HTTP_SIDE_BOTH);
   return(HTTP_ERR_POLICY);
  }
  if((x == Z_DENY) || (x == Z_REJECT))
  {
   http_log(s, HTTP_ERR_POLICY, NULL);
   http_close_imminent(s, HTTP_SIDE_SERVER);
   if(http_sendfile(s, HTTP_EFID_POLICY) == HTTP_ERROR)
   {
    http_log(s, HTTP_ERR_CLIENT_FAIL, NULL);
    return(HTTP_ERR_CLIENT_FAIL);
   }  
   return(HTTP_ERR_POLICY);
  }
 }
// 5.
 http_log(s, HTTP_EVENT_FILTER_OUT, tnewbuf->str);
 http_str_adds(tnewbuf, "\r\n"); //should be placed in the transfer filter
 if(http_write(s, HTTP_SIDE_SERVER, tnewbuf)==HTTP_ERROR)
 {
  http_log(s, HTTP_ERR_SERVER_FAIL, NULL);
  http_close_imminent(s, HTTP_SIDE_SERVER);
  if(http_sendfile(s, HTTP_EFID_SERVER) == HTTP_ERROR)
  {
   http_log(s, HTTP_ERR_CLIENT_FAIL, NULL);
   return(HTTP_ERR_CLIENT_FAIL);
  }
  return(HTTP_ERR_SERVER_FAIL);
 }

// 6.
 //defaults for data transfer
 s->filter_transfer= http_transfer_filter_copy;
 s->filter_encoding= http_encoding_filter_binary;
 s->filter_content=  http_type_filter_binary;
 y=0;   //=invalid exit code
 x=-2;  //data size (-1=unknown; -2=no data)

 if(http_lookup(tbuf, "Transfer-Encoding", t1, TRUE)==TRUE) //should be extended with "multipart/mime" !!!!
 {                                                                //&& (http_str_cmpis(t1, "chunked")==TRUE)
  s->filter_transfer=http_transfer_filter_chunked;                //chunked encoding
  x=-1;
 }else if(http_lookup(tbuf, "Content-Length", t1, TRUE)==TRUE)
 {
  if(sscanf(t1->str, "%u", &x)==0) break;
  //x=strtoul(tbuf->str, &tbp, 0);
  if(x==0) x=-2; //hack
  s->filter_transfer=http_transfer_filter_copy;                   //binary encoding
 }else
 {
  x=-2;                                                           //no data
 }
 
 if(x!=-2)
 {
  if(s->filter_transfer(s, HTTP_SIDE_CLIENT, HTTP_SIDE_SERVER, s->filter_encoding, s->filter_content, x, &y)==HTTP_ERROR)
  { 
   if(y==HTTP_SIDE_CLIENT)
   { 
    http_log(s, HTTP_ERR_CLIENT_FAIL, NULL);
    http_close_imminent(s, HTTP_SIDE_BOTH);
    http_filter_free(s);
    return(HTTP_ERR_CLIENT_FAIL);
   }else if(y==HTTP_SIDE_SERVER)
   {
    http_log(s, HTTP_ERR_SERVER_FAIL, NULL);
    http_close_imminent(s, HTTP_SIDE_BOTH);
    http_filter_free(s);
    return(HTTP_ERR_SERVER_FAIL);
   }else
   {
    http_log(s, HTTP_ERR_INTERNAL, NULL);
    http_close_imminent(s, HTTP_SIDE_BOTH);
    http_filter_free(s);   
    return(HTTP_ERR_INTERNAL);
   }
  }
  http_filter_free(s); //in case of normal operation this isn't needed
 }else
 {
  /* not needed as long as the request terminator is not in the transfer filter
  http_str_sets(tnewbuf, ""); //no data: terminate request
  if(http_writeline(s, HTTP_SIDE_SERVER, tnewbuf)==HTTP_ERROR)
  {
   http_log(s, HTTP_ERR_SERVER_FAIL, NULL);
   http_close_imminent(s, HTTP_SIDE_BOTH);
   return(HTTP_ERR_SERVER_FAIL); 
  } 
  */
 }

// 7.
 http_log(s, HTTP_EVENT_RESPONSE, NULL);
 do
 {
  http_str_sets(tbuf, ""); x=0; y=FALSE;
  while(x<(s->maxrespline))
  {
   if(http_readline(s, HTTP_SIDE_SERVER, t1, s->maxlinelen)==HTTP_ERROR)
   {
    http_close_imminent(s, HTTP_SIDE_SERVER);
    http_log(s, HTTP_ERR_SERVER_FAIL, NULL);
    if(http_sendfile(s, HTTP_EFID_SERVER) == HTTP_ERROR)
    {
     http_log(s, HTTP_ERR_CLIENT_FAIL, NULL);
     return(HTTP_ERR_CLIENT_FAIL);
    }
    return(HTTP_ERR_SERVER_FAIL);
   }
   if((t1->len)!=0) y=TRUE; //skip first empty lines
   if(((t1->len)==0) && (y==TRUE)) break;
   if(y==TRUE)
   {
    http_str_add(tbuf, t1);  
    http_str_adds(tbuf, "\n");
   }
   x++;
  }
  if((x>=(s->maxrespline)) || ((tbuf->len)==0))
  {
   http_close_imminent(s, HTTP_SIDE_SERVER);
   if(http_sendfile(s, HTTP_EFID_SERVER) == HTTP_ERROR)
   {
    http_log(s, HTTP_ERR_CLIENT_FAIL, NULL);
    return(HTTP_ERR_CLIENT_FAIL);
   }
   http_log(s, HTTP_ERR_SERVER_FAIL, NULL);
   return(HTTP_ERR_SERVER_FAIL);
  }
  http_str_copyn(t1, tbuf, http_str_contains_c(tbuf, '\n'));

  //=>bug: the http/0.9 didn't have this format (no response code and headers)
  if(http_splitresp(t1, t2, t3, t4)==FALSE) //t1:src 2:ver 3:code 4:text
  {
   http_log(s, HTTP_ERR_SERVER_SYNTAX, t1->str);
   if(http_sendfile(s, HTTP_EFID_SERVER)==HTTP_ERROR)
   {
    http_log(s, HTTP_ERR_CLIENT_FAIL, NULL);
    return(HTTP_ERR_CLIENT_FAIL);
   }
   return(HTTP_ERR_SERVER_SYNTAX);
  } 
  s->respflg=http_get_resp_param(t3->str); //get protocol dependent flags
  if((s->respflg)==-1) //unknown response
  {
   http_log(s, HTTP_ERR_SERVER_SYNTAX, t1->str);
   if(http_sendfile(s, HTTP_EFID_SERVER)==HTTP_ERROR)
   {
    http_log(s, HTTP_ERR_CLIENT_FAIL, NULL);
    return(HTTP_ERR_CLIENT_FAIL);
   }
   return(HTTP_ERR_SERVER_SYNTAX);
  }
 }while(http_str_cmpis(t3, "100")==TRUE); //skip 100 continue
 
 //check protocol version
 if(http_str_cmpis(t2, "HTTP/1.0")==TRUE)
 {
  s->mode=HTTP_MODE_CLOSE;
  s->ver=0x100;
 }else if((http_str_cmpis(t2, "HTTP/1.1")==TRUE) && ((s->mode)==HTTP_MODE_KEEPALIVE))
 {                                                  //^ hack for netscape & apache
  s->mode=HTTP_MODE_KEEPALIVE;
  s->ver=0x101;
 }else if((http_str_cmpis(t2, "HTTP/1.1")==TRUE) && ((s->mode)==HTTP_MODE_CLOSE))
 {                                                  //^ hack for netscape & apache
  s->mode=HTTP_MODE_CLOSE;
  s->ver=0x101;
 }else if(http_str_cmpis(t2, "")==TRUE)
 {
  s->mode=HTTP_MODE_CLOSE;
  s->ver=0x009;
 }else
 {
  http_log(s, HTTP_ERR_SERVER_SYNTAX, NULL);
  if(http_sendfile(s, HTTP_EFID_SYNTAX) == HTTP_ERROR)
  {
   http_log(s, HTTP_ERR_CLIENT_FAIL, NULL);
   return(HTTP_ERR_CLIENT_FAIL);
  }
  return(HTTP_ERR_SERVER_SYNTAX);
 }
 //check protocol override
 if(http_lookup(tbuf, "Connection", t1, TRUE)==TRUE)
 {
  if(http_str_cmpis(t1, "Keep-Alive")==TRUE) s->mode=HTTP_MODE_KEEPALIVE; else
  if(http_str_cmpis(t1, "Close")==TRUE)      s->mode=HTTP_MODE_CLOSE; else
  s->mode=HTTP_MODE_CLOSE;
 }
 if(http_lookup(tbuf, "Proxy-Connection", t1, TRUE)==TRUE)
 {
  if(http_str_cmpis(t1, "Keep-Alive")==TRUE) s->mode=HTTP_MODE_KEEPALIVE; else
  if(http_str_cmpis(t1, "Close")==TRUE)      s->mode=HTTP_MODE_CLOSE; else
  s->mode=HTTP_MODE_CLOSE;
 }
 
 //filter
 http_log(s, HTTP_EVENT_FILTER_IN, tbuf->str);
 http_str_sets(tnewbuf, "");
// 8.
 s->respflg=http_get_resp_param(t3->str);
 if(((s->reqflg) & HTTP_RESP_FLG_STOP) != 0)
 {
  http_log(s, HTTP_ERR_SERVER_SYNTAX, NULL);
  if(http_sendfile(s, HTTP_EFID_SYNTAX) == HTTP_ERROR)
  {
   http_log(s, HTTP_ERR_CLIENT_FAIL, NULL);
   return(HTTP_ERR_CLIENT_FAIL);
  }
  return(HTTP_ERR_SERVER_SYNTAX);
 }
 http_str_add(tnewbuf, t2); http_str_adds(tnewbuf, " ");
 http_str_add(tnewbuf, t3); http_str_adds(tnewbuf, " ");
 http_str_add(tnewbuf, t4);
 http_str_adds(tnewbuf, "\r\n");

 x=http_str_contains_c(tbuf, '\n'); x++; //skip req (if(x==-1) => x=0)
 http_str_sets(t1, "");
 while((tbuf->str)[x]!=0)
 {
  if((tbuf->str)[x]=='\n')
  {
   z_policy_lock(s->super.thread);
   if(http_filter(s, HTTP_FILTER_WHAT_RESP_HEAD, t1, NULL)==TRUE)
   {
    http_str_adds(t1, "\r\n");
    http_str_add(tnewbuf, t1);
    http_str_sets(t1, "");
   }
   z_policy_unlock(s->super.thread);
  }
  else
  {
   http_str_addc(t1, (tbuf->str)[x]);
  }
  x++;
 }

// 9.
 http_log(s, HTTP_EVENT_FILTER_OUT, tnewbuf->str);
 http_str_adds(tnewbuf, "\r\n"); //should be moved into the transfer filter
 if(http_write(s, HTTP_SIDE_CLIENT, tnewbuf)==HTTP_ERROR)
 {
  http_log(s, HTTP_ERR_CLIENT_FAIL, NULL);
  http_close_imminent(s, HTTP_SIDE_BOTH);
  return(HTTP_ERR_CLIENT_FAIL);
 }

// 10.
 if(((((s->reqflg)  & HTTP_REQ_FLG_HEAD)  != 0) && (((s->respflg) & HTTP_RESP_FLG_DATA_COND) != 0)) || //head & data_cond OR
    ((((s->respflg) & HTTP_RESP_FLG_DATA) == 0) && (((s->respflg) & HTTP_RESP_FLG_DATA_COND) == 0)) )  //!data & !data_cond
 { // if HEAD and data is dependent on it OR no data and no condicional data
  /* not needed as long as the terminator is not in the transfer filter
  http_str_sets(tnewbuf, ""); //termiante response
  if(http_writeline(s, HTTP_SIDE_CLIENT, tnewbuf)==HTTP_ERROR)
  {
   http_log(s, HTTP_ERR_CLIENT_FAIL, NULL);
   http_close_imminent(s, HTTP_SIDE_BOTH);
   return(HTTP_ERR_CLIENT_FAIL);
  }
  */
 }else // data must pass (if any)
 {
  //defaults for data transfer
  s->filter_transfer= http_transfer_filter_copy;
  s->filter_encoding= http_encoding_filter_binary;
  s->filter_content=  http_type_filter_binary;
  y=0;   //exit code
  x=-1;  //data size (-1=unknown; -2=no data)
 
  if(http_lookup(tbuf, "Transfer-Encoding", t1, TRUE)==TRUE)
  {
   s->filter_transfer=http_transfer_filter_chunked;                      //chunked encoding
   x=-1;
   z_debug(7, "USING chunked");
  }else if(http_lookup(tbuf, "Content-Length", t1, TRUE)==TRUE)
  {
   if(sscanf(t1->str, "%u", &x)==0) break;
   //x=strtoul(tbuf->str, &tbp, 0);
   //if(x==0) x=-2; //hack for bogus cgi-s!!!
   s->filter_transfer=http_transfer_filter_copy;                         //binary encoding
   z_debug(7, "USING sized binary");
  }else
  {
   x=-1;
   s->filter_transfer=http_transfer_filter_copy;
   s->mode=HTTP_MODE_CLOSE;                                              //binary encoding with close as eof
   z_debug(7, "USING copy to close");
  }
 
  if(x!=-2)
  {
   if(s->filter_transfer(s, HTTP_SIDE_SERVER, HTTP_SIDE_CLIENT, s->filter_encoding, s->filter_content, x, &y)==HTTP_ERROR)
   { 
    if(y==HTTP_SIDE_CLIENT)
    { 
     http_log(s, HTTP_ERR_CLIENT_FAIL, NULL);
     http_close_imminent(s, HTTP_SIDE_BOTH);
     http_filter_free(s);
     return(HTTP_ERR_CLIENT_FAIL);
    }else if(y==HTTP_SIDE_SERVER)
    {
     http_log(s, HTTP_ERR_SERVER_FAIL, NULL);
     http_close_imminent(s, HTTP_SIDE_BOTH);
     http_filter_free(s);
     return(HTTP_ERR_SERVER_FAIL);
    }else
    {
     http_log(s, HTTP_ERR_INTERNAL, NULL);
     http_close_imminent(s, HTTP_SIDE_BOTH);
     http_filter_free(s);   
     return(HTTP_ERR_INTERNAL);
    } 
   }
   http_filter_free(s); //in case of normal operation this isn't needed
  }
 }
 if(((s->mode)==HTTP_MODE_CLOSE) || (((s->respflg) & HTTP_RESP_FLG_CLOSE)!=0)) break;
}
//11.
 http_close_imminent(s, HTTP_SIDE_SERVER);
 http_close_graceful(s, HTTP_SIDE_CLIENT);
//12. 
 http_log(s, HTTP_EVENT_COMPLETE, NULL);
 return(HTTP_EVENT_COMPLETE);
}

//----------------------------------------------------------------

void http_init(void* selfp)
{
 HttpProxy *self = (HttpProxy*) selfp;
// ZPolicyObj *res;
 HttpStr tb; HttpStr tnb; HttpStr t1; HttpStr t2; HttpStr t3; HttpStr t4; HttpStr t5; HttpStr t6;
 HttpStr st1; HttpStr st2;

 self->cstat=TRUE;           //defaults ...
 self->sstat=FALSE;          //set it to TRUE after connect
 self->mode=HTTP_MODE_CLOSE; //connection close
 self->transp=FALSE;         //not transparent
 self->conv2proxy=FALSE;     //no parent proxy
 self->ver=0;
 self->maxreqline=50;        //seems like a good setup ?!?
 self->maxrespline=50;
 self->maxlinelen=2048;
 self->reqflg=0;
 self->respflg=0;

 self->req_cmd_def=HTTP_FILTER_DROP;   //GET, HEAD, POST added at setup (if not overrided)
 self->req_head_def=HTTP_FILTER_PASS;
 self->resp_cmd_def=HTTP_FILTER_PASS;
 self->resp_head_def=HTTP_FILTER_PASS;

 self->filter_transfer=      NULL;
 self->filter_encoding=      NULL;
 self->filter_content=       NULL;
 self->filter_transfer_data= NULL;
 self->filter_encoding_data= NULL;
 self->filter_content_data=  NULL;
 self->filter_transfer_error=NULL;
 self->filter_encoding_error=NULL;
 self->filter_content_error= NULL;    

 self->bufcpos=0;
 self->bufclen=0;
 self->bufspos=0;
 self->bufslen=0;
 self->logmode=HTTP_LOG_ERR0 | HTTP_LOG_ERR1 | HTTP_LOG_ERR2 | HTTP_LOG_ERR3 | 
               HTTP_LOG_EVN0 | HTTP_LOG_EVN1 | HTTP_LOG_EVN2 | HTTP_LOG_EVN3;
            
 self->addr = &st1;
 self->old_addr = &st2;
 self->transp_addr = FALSE;
 
 http_str_init(&st1); http_str_init(&st2);
 http_str_init(&tb); http_str_init(&tnb); 
 http_str_init(&t1); http_str_init(&t2); http_str_init(&t3); 
 http_str_init(&t4); http_str_init(&t5); http_str_init(&t6);
 
 self->req_cmd=              g_hash_table_new((GHashFunc) http_filter_hash_bucket, (GCompareFunc) http_filter_hash_compare);
 self->req_head=             g_hash_table_new((GHashFunc) http_filter_hash_bucket, (GCompareFunc) http_filter_hash_compare);
 self->resp_cmd=             g_hash_table_new((GHashFunc) http_filter_hash_bucket, (GCompareFunc) http_filter_hash_compare);
 self->resp_head=            g_hash_table_new((GHashFunc) http_filter_hash_bucket, (GCompareFunc) http_filter_hash_compare);
 self->filter_encoding_hash= g_hash_table_new((GHashFunc) http_filter_hash_bucket, (GCompareFunc) http_filter_hash_compare);
 self->filter_content_hash=  g_hash_table_new((GHashFunc) http_filter_hash_bucket, (GCompareFunc) http_filter_hash_compare); 
 
 //Z_VAR_TYPE_INT | Z_VAR_TYPE_HASH | Z_VAR_TYPE_STRING  | Z_VAR_GET | Z_VAR_SET  | Z_VAR_GET_CONFIG | Z_VAR_SET_CONFIG 
 z_proxy_var_new (&self->super, "connection_status_client", Z_VAR_TYPE_INT | Z_VAR_GET, &self->cstat);
 z_proxy_var_new (&self->super, "connection_status_server",   Z_VAR_TYPE_INT | Z_VAR_GET, &self->sstat);
 z_proxy_var_new (&self->super, "connection_mode",  Z_VAR_TYPE_INT | Z_VAR_GET, &self->mode);
 z_proxy_var_new (&self->super, "transparent_mode", Z_VAR_TYPE_INT | Z_VAR_GET | Z_VAR_SET_CONFIG, &self->transp);
 z_proxy_var_new (&self->super, "convert_to_proxy_request", Z_VAR_TYPE_INT | Z_VAR_GET | Z_VAR_SET_CONFIG, &self->conv2proxy);
 z_proxy_var_new (&self->super, "protocol_version", Z_VAR_TYPE_INT | Z_VAR_GET, &self->ver);
 z_proxy_var_new (&self->super, "log_mode",         Z_VAR_TYPE_INT | Z_VAR_GET | Z_VAR_SET_CONFIG, &self->logmode);
 z_proxy_var_new (&self->super, "max_req_line",     Z_VAR_TYPE_INT | Z_VAR_GET | Z_VAR_SET_CONFIG, &self->maxreqline);
 z_proxy_var_new (&self->super, "max_resp_line",    Z_VAR_TYPE_INT | Z_VAR_GET | Z_VAR_SET_CONFIG, &self->maxrespline);
 z_proxy_var_new (&self->super, "max_line_length",  Z_VAR_TYPE_INT | Z_VAR_GET | Z_VAR_SET_CONFIG, &self->maxlinelen);

 z_proxy_var_new (&self->super, "req_cmd_default",   Z_VAR_TYPE_INT | Z_VAR_GET | Z_VAR_SET_CONFIG, &self->req_cmd_def);
 z_proxy_var_new (&self->super, "req_head_default",  Z_VAR_TYPE_INT | Z_VAR_GET | Z_VAR_SET_CONFIG, &self->req_head_def);
 z_proxy_var_new (&self->super, "resp_cmd_default",  Z_VAR_TYPE_INT | Z_VAR_GET | Z_VAR_SET_CONFIG, &self->resp_cmd_def);
 z_proxy_var_new (&self->super, "resp_head_default", Z_VAR_TYPE_INT | Z_VAR_GET | Z_VAR_SET_CONFIG, &self->resp_head_def);
 
 z_proxy_var_new (&self->super, "request_commands",  Z_VAR_TYPE_HASH | Z_VAR_GET | Z_VAR_GET_CONFIG, self->req_cmd);
 z_proxy_var_new (&self->super, "request_headers",   Z_VAR_TYPE_HASH | Z_VAR_GET | Z_VAR_GET_CONFIG, self->req_head);
 z_proxy_var_new (&self->super, "response_commands", Z_VAR_TYPE_HASH | Z_VAR_GET | Z_VAR_GET_CONFIG, self->resp_cmd);
 z_proxy_var_new (&self->super, "response_headers",  Z_VAR_TYPE_HASH | Z_VAR_GET | Z_VAR_GET_CONFIG, self->resp_head);
 z_proxy_var_new (&self->super, "filter_encoding",   Z_VAR_TYPE_HASH | Z_VAR_GET | Z_VAR_GET_CONFIG, self->filter_encoding_hash);
 z_proxy_var_new (&self->super, "filter_content",    Z_VAR_TYPE_HASH | Z_VAR_GET | Z_VAR_GET_CONFIG, self->filter_content_hash);

 //z_proxy_var_new (&self->super, "server_address", Z_VAR_TYPE_STRING | Z_VAR_GET, &(self->addr->str)); // this isn't a GString!!!
 z_proxy_var_new (&self->super, "server_address_transparent", Z_VAR_TYPE_INT | Z_VAR_GET | Z_VAR_SET_CONFIG, &self->transp_addr);
  
 ZPROXY_SET_STATE(self, PS_CONFIG); 
 z_proxy_config_event(&self->super);
 
 ZPROXY_SET_STATE(self, PS_STARTING_UP);
 z_proxy_startup_event(&self->super);

 ZPROXY_SET_STATE(self, PS_WORKING);
 http_main(self ,&tb, &tnb, &t1, &t2, &t3, &t4, &t5, &t6); 

 ZPROXY_SET_STATE(self, PS_SHUTTING_DOWN);
 z_proxy_shutdown_event(&self->super);

 http_str_free(&st1); http_str_free(&st2);
 http_str_free(&tb); http_str_free(&tnb);
 http_str_free(&t1); http_str_free(&t2); http_str_free(&t3); 
 http_str_free(&t4); http_str_free(&t5); http_str_free(&t6);
 
 z_proxy_destroy(&self->super);  
 return;
}

ZProxy *http_proxy_new(gchar *session_id, ZStream *client, ZPolicyObj *handler)
{
  HttpProxy *self = g_new0(HttpProxy, 1);
  char buf[512];

  z_proxy_init(&self->super, session_id, client, handler);
  g_snprintf(buf, sizeof(buf), "http proxy thread (%s)", client->name);
  z_thread_new(buf, http_init, self);
  return(&self->super);
}

gint zorp_module_init(void)
{
  z_registry_add("http", ZR_PROXY, http_proxy_new);
  return(TRUE);
}
