############################################################################
##
## Copyright (c) 2000 BalaBit IT Ltd, Budapest, Hungary
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this program; if not, write to the Free Software
## Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
##
## $Id: Chainer.py,v 1.20 2000/10/10 18:25:59 bazsi Exp $
##
## Author  : Bazsi
## Auditor : kisza
## Last version: 1.18
## Notes:
##   - comments in English
##
## Author  : Bazsi
## Auditor : kisza
## Last version: 1.18
## Notes:
##   - comments in English
##
############################################################################

from Zorp import *
import Connector, Stream, SockAddr
import socket

class Chainer:
	"""
		Ez az osztaly hvdik meg, ha valaki hozzcsatlakozik a proxy-hoz,
		itt pl fel a kapcsolat a szerver oldalal.
		---
		Whenever someone connects to the proxy, this class will be
		invoked. The connection with the server side will be
		constructed here.
	"""
	def connectServer(self, session, host):
		raise NotImplementedError
		
	def establishConnection(self, session, local, remote):
		"""
			Ez a fggvny hajtja vgre a csatlakozst, ahhoz a serverhez, amit a
			connectServer fgv. meghatrozott
			Paramterek:
				remote			A server adatai, amihez csatlakozni kell.
			Lers:
				Ez a fggvny ltrehozza a csatlakozst a proxy-tol a
				megadott szerverig.
			Visszatrsi rtk:
				int				A szerver fel men csatorna fd-je.
			---
			This function does the connection to the server determined by
			the connectServer method.
			Parameters:
				remote			The data of the server
			Workings:
				This function creates the connection from the proxy to the server.
			Return value:
				int				The fd of the connection to the server
		"""
		try:
			session.setServer(remote)
			conn = Connector.Connect(local, remote)
			session.server_fd = conn.blockingConnect()
			if session.server_fd != -1:
				session.server = Stream.Stream(session.server_fd, session.session_id + ": server")
				return session.server_fd
		except DACException:
			pass
		except MACException:
			pass
		return -1

		

class DirectedChainer(Chainer):
	"""
		Ez a csatlakozs elre meghatrozott szerverhez csatlakozik,
		brmely krs esetn

		local						A csatlkozs loklis adatai
		remote					A szerver adatai
		---
		This chainer connects to a predefined server, in any case.
		local			The local data of the connection
		remote			The data of the server
	"""
	local = None
	remote = None
	
	def __init__(self, local, remote):
		"""
			Ltrehoz egy Chainer objektumot.

			Paramterek:
				local				A csatlakozs loklis adatai
				remote			A szerver adatai, amihez csatlakozni kell

			Lers:
				Belltja a loklis vltozkat a megadott paramterek
				alapjn

			Visszatrsi rtk:
				-
			---
			Creates a Chainer object.
			Parameters:
				local			The local data of the connection
				remote			The data of the server
			Workings:
				Sets the instance attributes by the given data
			Return value:
				-
		"""
		self.local = local
		self.remote = remote
	
	def connectServer(self, session, host, port):
		"""
			Ez a fuggvny hvodik meg kvlrl csatlakozs esetn, s dnti el,
			hogy hova is kell a proxy-nak csatlakoznia

			Paramterek:
				session			Az aktulis kapcsols adatai
				host				Az tadott gepcm, ??Nincs hasznlatban.??
				port				Az tadott portcm. ??Nincs hasznlatban.??

			Lers:
				Elinditja az establishConnection-t a beltott rtkeknek
				megfelelen.

			Visszatrsi rtk:
				int				Az establishConnection lltal visszaadott fd.
			---
			This function is called in case a connection comes, and decides where the proxy
			should connect

			Parameters:
				session			The data of the current session
				host			dummy parameter
				port			dummy parameter
			Return value:
				int			The fd given by establishConnection
		"""
		return self.establishConnection(session, self.local, self.remote)

class TransparentChainer(Chainer):
	"""
		Ez a csatlakozs kapcsolatonknt ms, a kapcsolat lltal meghatrozott
		helyhez csatlakozik.

		Vltozk:
			forge_addr			trja-e a kliens adatait egy megadottra
									csatlakozskor
			local					Ha igen, akkor mire.
		---
		This chainer connects to the destination determined by the client
		Attributes:
			forge_addr			if the client address shall be forged
			local				if yes, what will be set as client address
	"""
	forge_addr = 0
	local = None
	
	def __init__(self, local, forge_addr):
		"""
			Ltrehozza az osztly egy pldnyt.

			Paramterek:
				local			Mire rja t a kliens adatait
				forge_addr		trja-e a kliens adatait

			Lers:
				Ltrehozza az osztly egy pldnyt, s a paramternek megfelelen
				belltja, hogy trdjanak-e a klien adatai

			Visszatrsi rtk:
				-
			---
			Creates an instance of the class
			Parameters:
				forge_addr			if the client address shall be forged
				local				if yes, what will be set as client address
			Workings:
				Creates an instance, sets its attributes
		FIXME:!!! self.local	= local #ejnye kisza
		"""
		self.forge_addr = forge_addr
	
	def connectServer(self, session, host, port):
		"""
			Ez a fuggvny hvodik meg kvlrl csatlakozs esetn, s dnti el,
			hogy hova is kell a proxy-nak csatlakoznia

			Paramterek:
				session			Az aktulis kapcsols adatai
				host				Az tadott gpcm. ??Nincs hasznlatban??
				port				Az tadott portcm. ??Nincs hasznlatban??

			Lers:
				Csatlakozik a session alltal meghatarozott helyre, opcionlisan
				elrejtve a kliens adatait.

			Visszatrsi rtk:
				int				Az establishConnect lltal visszaadott fd.
			---
			This function is called in case a connection comes, and decides where the proxy
			should connect

			Parameters:
				session			The data of the current session
				host			dummy parameter
				port			dummy parameter
			Return value:
				int			The fd given by establishConnection
		"""

		if self.forge_addr:
			local = session.client_addr
		else:
			local = self.local
		remote = getTransparentDestination(session.client_fd)
		return self.establishConnection(session, self.local, remote)

class InbandChainer(Chainer):
	"""
		Ezt a chainert hasznljk a proxyk addicionlis tcp kapcsolatok kialaktshoz
		pl ftp data csatorna

		Vltozk:
			local				A kapcsolat itteni vgnek a cme
		---
		This chainer is used by the proxies to establish additional tcp connections,
		like ftp data connecion
		Attributes:
			local				the address of the local end of the connection
	"""
	local = None
	
	def __init__(self, local):
		"""
			Ltrehozza az osztly egy pldnyt.

			Paramterek:
				local			Mire lltsa be a kliens adatait

			Lers:
				Ltrehozza az osztly egy pldnyt, s a paramternek megfelelen
				belltja, a loklis cmet

			Visszatrsi rtk:
				-
			---
			Creates an instance of the class
			Parameters:
				local				local address
			Workings:
				Creates an instance, sets its attribute
			Return value:
				-
			"""
		self.local = local
	
	def connectServer(self, session, host, port):
		"""
			Ezt a fggvnyt hvja a proxy ha csatlakozni akar
			Paramterek:
				session			Az aktulis kapcsols adatai
				host				A clcm
				port				A clport

			Lers:
				Csatlakozik a session alltal meghatarozott helyre

			Visszatrsi rtk:
				int				Az establishConnect lltal visszaadott fd.
			---
			This function is called by the proxy for a new connection
			Parameters:
				session			The data of the current session
				host			The address of the destioation
				port			The port of the destination
			Return value:
				int			The fd given by establishConnection
		"""
		
		#message(0, "InbandChainer.connectServer()")
		try:
			ip = socket.gethostbyname(host)
			return self.establishConnection(session, self.local, SockAddr.SockAddrInet(ip, port))
		except socket.error:
			return -1

class FailoverChainer(Chainer):
	"""
		Failover chainer: tbb cmet is megprbl egyms utn, amg egy ssze nem jn

		Vltozk:
			local				A kapcsolat itteni vgnek a cme
			hosts				A megprbland cmek listja
			current_host			Az ppen prblt cm indexe a listban
		---
		Failover chainer: tries more addresses in turn until one connection establishes
		Attributes:
			local				the address of the local end of the connection
			hosts				The list of addresses to try
			current_host			The index of the actual host in the list
	"""
	
	local = None
	hosts = []
	current_host = 0
	
	def __init__(self, local, hosts):
		"""
			Ltrehozza az osztly egy pldnyt.

			Paramterek:
				local			Mire lltsa be a kliens adatait
				hosts			A megprbland cmek listja

			Lers:
				Ltrehozza az osztly egy pldnyt, s a paramternek megfelelen
				belltja az attribtumokat

			Visszatrsi rtk:
				-
			---
			Creates an instance of the class
			Parameters:
				local			local address
				hosts			The list of addresses to try
			Workings:
				Creates an instance, sets its attributes
			Return value:
				-
		"""
			
		self.local = local
		self.hosts = hosts
	
	def nextHost(self):
		"""
			Megkeresi a listban a kvetkez hostot amit meg kell prblni
			Paramterek:
				-
			Lers:
				Egyet ad az indexhez (current_host), nullzza ha tlment.
			Visszatrsi rtk:
				-
			---
			Finds the next host to try
			Parameters:
				-
			Workings:
				Increments the index (current_host), zeroes it if overflowed
			Return value:
				-
		"""
		self.current_host = self.current_host + 1
		if self.current_host == len(self.hosts):
			self.current_host = 0

	def getHostAddress(self):
		"""
			Visszaadja az aktulis hostot
			Paramterek:
				-
			Lers:
				Visszaadja a lista (hosts) aktulis (current_host) elemt
			Visszatrsi rtk:
				-
			---
			Returns the actual host to try
			Parameters:
				-
			Workings:
				Reurns the actual (current_host) elem of the list (hosts)
			Return value:
				-
		"""
		return self.hosts[self.current_host]
	
	def connectServer(self, session, host, port):
		"""
			Ez a fuggvny hvodik meg kvlrl csatlakozs esetn, s dnti el,
			hogy hova is kell a proxy-nak csatlakoznia

			Paramterek:
				session			Az aktulis kapcsols adatai
				host				Az tadott gepcm, ??Nincs hasznlatban.??
				port				Az tadott portcm. ??Nincs hasznlatban.??

			Lers:
				Prblkozik az aktulis hosttal, s ha nem sikerl akkor krbemegy a 
				listn s mindent megprbl. Nem adja fel.
			Visszatrsi rtk:
				int				Az establishConnection lltal visszaadott fd.
			---
			This function is called in case a connection comes, and decides where the proxy
			should connect

			Parameters:
				session			The data of the current session
				host			dummy parameter
				port			dummy parameter
			Workings:
				Tries the actual host, if it does not succeeds, tries all the hosts round the
				list. It never gives up.
			Return value:
				int			The fd given by establishConnection
		"""
		
		first_attempt = self.current_host
		remote = self.getHostAddress()
		res = self.establishConnection(session, self.local, remote)
		self.nextHost()
		while res == -1 and self.current_host != first_attempt:
			remote = self.getHostAddress()
			res = self.establishConnection(session, self.local, remote)
			self.nextHost()
		return res
