############################################################################
##
## Copyright (c) 2000 BalaBit IT Ltd, Budapest, Hungary
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this program; if not, write to the Free Software
## Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
##
## $Id: Ftp.py,v 1.18 2000/10/10 18:30:57 bazsi Exp $
##
## Author  : kisza
## Auditor :  
## Last version: 
## Notes:
##
############################################################################

import Zorp, Service, Connector, Proxy, Session, Stream, SockAddr, Plug
from Zorp import *

#
# Mode flag bitmap
#
# rfc compliancy
rfc1545		=	0x80000000
rfc2228		=	0x40000000
rfc1579		=	0x20000000
rfc2389		=	0x10000000
rfc2428		=	0x08000000
rfc765		=	0x04000000
rfc775		=	0x02000000
rfc959		=	0x01000000
# functional command groups
levelConnect	=	0x00800000
levelLogin	=	0x00400000
levelDataprot	=	0x00200000
levelCommandc	=	0x00100000
levelLogoff	=	0x00080000
levelTranspar	=	0x00040000
levelInform	=	0x00020000
levelMisc	=	0x00010000
levelDirectory	=	0x00008000
levelFileact	=	0x00004000
levelMail	=	0x00002000
levelUnknown	=	0x00001000
# access types
accessRead	=	0x00000800
accessWrite	=	0x00000400
accessCtrlsys	=	0x00000200
accessCtrlnet	=	0x00000100
accessUnkn	=	0x00000080
# is the username known?
nameKnown	=	0x00000040
nameUnknown	=	0x00000020
nameReset	=	0x00000010
# command categories by information flow type
ccRead		=	0x00000008
ccWrite		=	0x00000004
ccControll	=	0x00000002
ccFeedback	=	0x00000001

#
# Predefinied modes
#
AnonymousRead	=       0x01BFFFFB
AllowAllSTD9	=	0x01FFFFFF

class FtpDataConnect:
	"""
		Helper class to handle data connection
		Attributes:
			state		The state of the connection
					States:
						0	- "0. side" connection establishment
						1	- "1.side" connection establishment
						2	connections are set up, stacking the proxy
			parms		Parameters for state 0 and 1
					It is a 2-element list of triples. The elements of the triples:
						0.:	Type of the connection
		 					L - listen
		 					C - connect
		 					A - 0. side listens with a precreated listener
		 					B - 1. side listens with a precreated listener
						1.:	address of the local end of the connection
						2.:	address of the remote end of the connection
			listen		Zorp.Listen object for type C FIXME why just one?
			listens		list of Zorp.Listen objects for 0. and 1. sides
			connect		Zorp.Connect object for type C FIXME isn't there a connect-connect case?
			fds		File descriptors for the sides
			session		Session object for the session
			data_proxy	The proxy class which should be stacked on the connection
	"""
        state = 0
        parms = [(), ()];
        listen = None
        connect = None
	listens = [None, None]
        ftp_proxy = None
        fds = [-1, -1]
        session = None

        def __init__(self, session, ftp_proxy):
 		"""
 			Constructor
 			Parameters:
 				session		see above
 				data_proxy	see above
 			Workings:
 				stores the parameters into attributes
 			Return value:
 				-
 		"""
        	self.session = session
        	self.ftp_proxy = ftp_proxy
        	
        def __del__(self):
 		"""
 			Destructor
 			Parameters:
 				-
 			Workings:
 				calls resetState
 			Return value:
 				-
 		"""
        	self.resetState()

	def resetState(self):
 		"""
 			Sets the initial state of everything
 			Parameters:
 				-
 			Workings:
 				Initialises attributes to none, destroys listeners and connectors
 			Return value:
 				-
 		"""
		if self.listen != None:
	        	self.listen.destroy()
	        if self.connect != None:
	        	self.connect.destroy()
	        if self.listens[0] != None:
	       		self.listens[0].destroy()
	       	if self.listens[1] != None:
	        	self.listens[1].destroy()
		self.listens = [None, None]
		self.listen = None
		self.connect = None
		self.state = 0

	def prepare(self, ip, side):
		"""
			Puts a (new) listener to the side

			Parameters:
				ip	The address where we listen
				side	The side
			Workings:
				If there were a listener for the side, it is destroyed
				Tries to listen on the given ip on any port.
				If gets an IOError exception it actually rises an AttributeError (see FIXME)
			Return value:
				The port where the Listener is bound

		"""
		if (side < 0 or side > 1):
			return 0
			
		if self.listens[side] != None:
			self.listens[side].destroy()
		
		try:
#			self.listens[side] = \
#				Zorp.ListenPortRange(SockAddr.SockAddrInet(ip, 0), 
#						     self.ftp_proxy.data_port_min,
#						     self.ftp_proxy.data_port_max,
#						     self.acceptedCallback, 
#						     1)
			self.listens[side] = \
				Zorp.Listen(SockAddr.SockAddrInet(ip, 0), 
					    self.acceptedCallback, 
					    1)
			return self.listens[side].local.port
		except IOError:
			self.listens[side] = None
			
		return 0

	def start(self, fst_mode, fst_local, fst_remote,
	                snd_mode, snd_local, snd_remote):
 		"""
 			Starts the state machine
 			Parameters:
 			{fst,snd}_	{First(0.),Second(1.)} side
 				mode	see attribute parms
 				local	-+-
 				remote	-+-
 			Workings:
 				sets attribute parms, zeroes out the state, and calls processState
 			Return value:
 				-
 		"""

                self.parms = [(fst_mode, fst_local, fst_remote),
                	      (snd_mode, snd_local, snd_remote)]
		self.state = 0
                self.processState()
                
        def processState(self):
		"""
			State machine
			Parameters:
				-
			Workings:
				If state == 2, stacks in the proxy and starts it with a session 
					stacked out from self.session
				In state 0 and 1 it either listens or connects, based on the type field
					of parms[state]
			Return value:
				-
		"""
		Zorp.debug(0, "data listener, state %d" % (self.state))
        	if self.state == 2:
        		# start proxy
			Zorp.debug(0, "stacking data proxy")
			ss = Session.StackedSession(self.session)
			ss.client_fd = self.fds[0]
			ss.client = Stream.Stream(ss.client_fd, ss.session_id + 'data client')
			ss.server_fd = self.fds[1]
			ss.server = Stream.Stream(ss.client_fd, ss.session_id + 'data server')
			self.ftp_proxy.data_proxy(ss)
        		return
        		
                parms = self.parms[self.state]
                if parms[0] == 'L':
			Zorp.debug(0, "data listener, standard")
			if self.listen != None:
				self.listen.destroy()
			try:
	                        self.listen = Zorp.Listen(parms[1], self.acceptedCallback, 1)
	                except IOError:
	                	self.listen = None
			#FIXME what if we just got an IOError???
                        self.listen.start()
                elif parms[0] == 'C':
			Zorp.debug(0, "data connector")
			if self.connect != None:
				self.connect.destroy()
			try:
	                        self.connect = Zorp.Connect(parms[1], parms[2], self.connectedCallback)
			except IOError:
			#FIXME what if we just got an IOError???
				self.connect = None
			
                elif parms[0] == 'A':
			Zorp.debug(0, "data listener, A")
                	self.listens[0].start()
              	elif parms[0] == 'B':
			Zorp.debug(0, "data listener, B")
              		self.listens[1].start()

        def acceptedCallback(self, client, fd):
		"""
			Callback for established connection on the Listener
			Parameters:
				client		the client which has connected
				fd		the file descriptor for the connection
			Workings:
				stores the parameters, increases the state and continues with the state machine
				FIXME:	do we check if we got the connection from the right place?
			Return value:
				None
		"""
		Zorp.debug(0, "data listener, acceptedCallback")
        	self.fds[self.state] = fd
                self.state = self.state + 1
                self.processState()

        def connectedCallback(self, fd):
		"""
			Callback for established connection on the Connector
			Parameters:
				fd		the file descriptor for the connection
			Workings:
				stores the parameter, increases the state and continues with the state machine
			Return value:
				None
		"""
		Zorp.debug(0, "data listener, connectedCallback")
        	self.fds[self.state] = fd
                self.state = self.state + 1
                self.processState()


# 
# Default FTP proxy - deny all
#
class FtpProxy(Proxy.Proxy):
 	"""
 		Default ftp proxy class: Deny everything (but data connection establishment)
 		(It is good to inherit from this class)
 		Attributes:
 			data_connect	Object for the data connection
 	"""
	data_connect = None
	data_proxy = Plug.PlugProxy
	data_port_min = 4000;
	data_port_max = 6000;

	def __init__(self, session):
 		"""
 			Constructor
 			Parameters:
 				session		The session object
 			Workings:
 				initializes an ftp proxy and defines that
 				the data connection will flow through a plug using FtpDataConnect
 			Return value:
 				-
 		"""
		Proxy.Proxy.__init__(self, "ftp", session)
		self.data_connect = FtpDataConnect(session, self)
		#, self.data_proxy_min, self.data_proxy_max)

	def user(self, dir, uname):
 		"""
 			USER command
 			Parameters:
 				dir	information flow type
 				uname	user name
 			Workings:
 				denies it
 			Return value:
 				Z_DENY
 		"""
		#Zorp.debug(0, "FTPE: USER %s , flag %c" % (uname, dir))
		return Z_REJECT

	def password(self, dir, uname):
 		"""
 			PASS command
 			Parameters:
 				dir	information flow type
 				uname	user name
 			Workings:
 				denies it
 			Return value:
 				Z_DENY
 		"""
		#Zorp.debug(0, "FTPE: PASS , flag %c" % (dir))
		return Z_REJECT

	def nontransparent(self, dir, dst, user, passw):
 		"""
 			non-transparent authentication
 			Parameters:
 				dir	information flow type
 				dst	destination
 				user	user name
 				passw	password
 			Workings:
 				denies it
 			Return value:
 				Z_DENY
 		"""
		#Zorp.debug(0, "FTPE: nont %s %s %s, flag %c" % (dst, user, passw, dir))
		return Z_REJECT

	def userAUTH(self, dir, uname):
 		"""
 			Authentication
 			Parameters:
 				dir	information flow type
 				uname	user name
 			Workings:
 				denies it
 			Return value:
 				Z_DENY
 		"""
		#Zorp.debug(0, "FTPE: AUTH %s , flag %c" % (uname, dir))
		return Z_REJECT

	def quit(self, dir):
 		"""
 			QUIT command
 			Parameters:
 				dir	information flow type
 			Workings:
 				denies it
			Return value:
 				Z_DENY
 		"""
		#Zorp.debug(0, "FTPE: QUIT flag %c" % (dir))
		return Z_REJECT

	def bounceCheck(self, dir, realip, realport, reqip, reqport):
 		"""
 			Bounce checking? FIXME could not find its C half
 			Parameters:
 				dir	information flow type
 				hashv
 				comm
 				param
 			Workings:
 				denies it
 			Return value:
 				Z_DENY
 		"""
		#Zorp.debug(0, "FTPE: bounce attack detected")
		return Z_REJECT

        def portCommand(self, dir, hashv, comm, param, ip, port):
 		"""
 			PORT command
 			Parameters:
 				dir	information flow type
 				hashv	hash value
 				comm	command
 				param	parameter
 				ip	ip address of requested data onnection
 				port	port of requested data onnection
 			Workings:
 				denies it
 			Return value:
 				Z_DENY
 		"""
                #Zorp.debug(0, "FTPE: command: %s %s hash: %X flag: %s" % ( comm, param, hashv, dir))
                #Zorp.debug(0, "FTPE: %s:%d" % (ip, port))
                return Z_REJECT

        def pasvAnswer(self, dir, line, ip, port):
 		"""
 			ansver to a PASV commend
 			Parameters:
 				dir	information flow type
 				line
 				ip	ip address of requested data onnection
 				port	port of requested data onnection
 			Workings:
				denies it
 			Return value:
 				Z_DENY
 		"""
                #Zorp.debug(0, "FTPE: %s flag: %s" % (line, dir))
                #Zorp.debug(0, "FTPE: %s:%d" % (ip, port))
                return Z_REJECT

        def answer(self, dir, line):
 		"""
 			answer
 			Parameters:
 				dir	information flow type
 				line	the answer itself
			Workings:
 				denies it
 			Return value:
 				Z_DENY
 		"""
                #Zorp.debug(0, "FTPE: %s flag: %s" % (line, dir))
                return Z_REJECT

	def transferCommand(self, dir, hashv, comm, param):
 		"""
 			transfer command
 			Parameters:
 				dir	information flow type
 				hashv	hash value
 				comm	command
 				param	parameter
 			Workings:
 				denies it
 			Return value:
 				Z_DENY
 		"""
		#Zorp.debug(0, "FTPE: command: %s %s hash: %X flag: %s" % ( comm, param, hashv, dir))
		return Z_REJECT

	def dataTransfer(self, dir, hashv, comm, param):
 		"""
 			data transfer
 			Parameters:
 				dir	information flow type
 				hashv	hash value
 				comm	command
 				param	parameter
 			Workings:
 				denies it
 			Return value:
 				Z_DENY
 		"""
		#Zorp.debug(0, "FTPE: command: %s %s hash: %X flag: %s" % ( comm, param, hashv, dir))
		return Z_REJECT

	def directoryChange(self, dir, hashv, comm, param):
 		"""
 			directory change command
 			Parameters:
 				dir	information flow type
 				hashv	hash value
 				comm	command
 				param	parameter
 			Workings:
 				denies it
 			Return value:
 				Z_DENY
 		"""
		#Zorp.debug(0, "FTPE: command: %s %s hash: %X flag: %s" % ( comm, param, hashv, dir))
		return Z_REJECT

	def directoryManipulation(self, dir, hashv, comm, param):
 		"""
 			directory manipulation command
 			Parameters:
 				dir	information flow type
 				hashv	hash value
 				comm	command
 				param	parameter
			Workings:
				denies it
 			Return value:
 				Z_DENY
 		"""
		#Zorp.debug(0, "FTPE: command: %s %s hash: %X flag: %s" % ( comm, param, hashv, dir))
		return Z_REJECT

	def siteCommand(self, dir, hashv, comm, param):
 		"""
 			SITE command
 			Parameters:
 				dir	information flow type
 				hashv	hash value
 				comm	command
 				param	parameter
			Workings:
 				denies it
 			Return value:
 				Z_DENY
 		"""
		#Zorp.debug(0, "FTPE: command: %s %s hash: %X flag: %s" % ( comm, param, hashv, dir))
		return Z_REJECT

	def mailCommand(self, dir, hashv, comm, param):
 		"""
 			mail command
 			Parameters:
 				dir	information flow type
 				hashv	hash value
 				comm	command
 				param	parameter
 			Workings:
 				denies it
 			Return value:
 				Z_DENY
 		"""
		#Zorp.debug(0, "FTPE: command: %s %s hash: %X flag: %s" % ( comm, param, hashv, dir))
		return Z_REJECT

	def miscCommand(self, dir, hashv, comm, param):
 		"""
 			miscellaneous command
 			Parameters:
 				dir	information flow type
 				hashv	hash value
 				comm	command
 				param	parameter
 			Workings:
				denies it
 			Return value:
 				Z_DENY
 		"""
		#Zorp.debug(0, "FTPE: command: %s %s hash: %X flag: %s" % ( comm, param, hashv, dir))
		return Z_REJECT

	def generalCommand(self, dir, hashv, comm, param):
		#Zorp.debug(0, "FTPE: command: %s %s hash: %X flag: %s" % ( comm, param, hashv, dir))
		return Z_REJECT

	def unknownCommand(self, dir, comm, param):
		#Zorp.debug(0, "FTPE: command: %s %s flag: %s" % ( comm, param, dir))
		return Z_REJECT

	def sendABOR(self, dir):
		#Zorp.debug(0, "FTPE: ABOR flag: %s" % ( dir))
		return Z_REJECT

	def prepareDATA(self, ip, side):
		return self.data_connect.prepare(ip, side)

	def startDATA(self, dir, fst_mode, fst_loc_addr, fst_loc_port, fst_rem_addr, fst_rem_port, snd_mode, snd_loc_addr, snd_loc_port, snd_rem_addr, snd_rem_port):
		#
		# mode-ok:
		#  L - listen
		#  C - connect
		#  A - 0. oldal elore foglalt porttal listen
		#  B - 1. oldal elore foglalt porttal listen
		#
		#Zorp.debug(0, "FTPE: Rem %s:%d - Loc %s:%d -%c" % (fst_rem_addr, fst_rem_port, fst_loc_addr, fst_loc_port, fst_mode))
		#Zorp.debug(0, "FTPE: Rem %s:%d - Loc %s:%d -%c" % (snd_rem_addr, snd_rem_port, snd_loc_addr, snd_loc_port, snd_mode))
		#Zorp.debug(0, "FTPE:  %s" % (dir))

		def convertSockAddr(addr, port):
			"""
				helper function to convert an ip and port into sockaddr
				Parameters:
					addr	address
					port	port
				Workings:
					if address is an empty string, returns None
					otherwise returns the sockaddr
				Return value:
					the converted sockaddr or None
			"""
			if addr == '':
				return None
			else:
				return SockAddr.SockAddrInet(addr, port)

		self.data_connect.start(fst_mode, 
					convertSockAddr(fst_loc_addr, fst_loc_port),
					convertSockAddr(fst_rem_addr, fst_rem_port),
					snd_mode,
					convertSockAddr(snd_loc_addr, snd_loc_port),
					convertSockAddr(snd_rem_addr, snd_rem_port))
		return Z_ACCEPT

#
#  Ftp proxy - allows anonymous download
#
class FtpProxyAnonymous(FtpProxy):
	"""
		Ftp proxy - allows anonymous download
		Attributes:
			-
	"""

	def startUp(self):
		"""
			Method called at proxy startup
			Parameters:
				-
			Workings:
				sets the set of accepted commands bitmap
			Return value:
				-
		"""
		self.command_accept = AnonymousRead

	def config(self):
		"""
			Method called at proxy config time
			Parameters:
				-
			Workings:
				sets the set of accepted commands bitmap
			Return value:
				-
		"""
		self.command_accept = AnonymousRead
		
	def user(self, dir, uname):
		"""
			USER command
			Parameters:
				dir	information flow type
				uname	user name
			Workings:
				Accepts only ftp or anonymous user
			Return value:
				Z_ACCEPT for accepted users, Z_REJECT otherwise
		"""
		if  ( uname == "anonymous" or uname == "ftp" ):
			return Z_ACCEPT
		else:
			return Z_REJECT

	def password(self, dir, uname):
		"""
			PASS command
			Parameters:
				dir	information flow type
				uname	user name
			Workings:
				Accept everything
			FIXME: how we set the outgoing password to some value?
			FIXME: don't we need a password parameter?
		"""
		return Z_ACCEPT

	def nontransparent(self, dir, dst, user, passw):
		"""
			non-transparent login
			Parameters:
				dir	information flow type
				dst	destination
				uname	user name
				passw	password
			Workings:
				Accepts only ftp or anonymous user
			Return value:
				Z_ACCEPT for accepted users, Z_REJECT otherwise
		"""
		if  ( uname == "anonymous" or uname == "ftp" ):
			return Z_ACCEPT
		else:
			return Z_REJECT

        def portCommand(self, dir, hashv, comm, param, ip, port):
		"""
			PORT command
			Parameters:
				dir	information flow type
				hashv	hash value
				comm	command
				param	parameter
				ip	ip address of requested data onnection
				port	port of requested data onnection
			Workings:
				accepts it
			Return value:
				Z_ACCEPT
		"""
                return Z_ACCEPT

        def pasvAnswer(self, dir, line, ip, port):
		"""
			ansver to a PASV commend
			Parameters:
				dir	information flow type
				line
				ip	ip address of requested data onnection
				port	port of requested data onnection
			Workings:
				accepts it
			Return value:
				Z_ACCEPT
		"""
                return Z_ACCEPT

	def unknownCommand(self, dir, comm, param):
		"""
			unknown command
			Parameters:
				dir	information flow type
				hashv	hash value
				comm	command
				param	parameter
			Workings:
				accepts it
			Return value:
				Z_ACCEPT
		"""
		return Z_ACCEPT

#
# Ftp proxy - R U kidding? Allow all
#
class FtpProxyAllow(FtpProxy):
	"""
		An ftp proxy which allows everything defined in RFC 959
		DO NOT USE FOR any sensible setup!!
		Attributes:
			-
	"""

	def startUp(self):
		"""
			Method called at proxy startup
			Parameters:
				-
			Workings:
				sets the set of accepted commands bitmap
			Return value:
				-
		"""
		self.command_accept = AllowAllSTD9

	def config(self):
		"""
			Method called at proxy config time
			Parameters:
				-
			Workings:
				sets the set of accepted commands bitmap
			Return value:
				-
		"""
		self.command_accept = AllowAllSTD9
		
        def portCommand(self, dir, hashv, comm, param, ip, port):
		"""
			PORT command
			Parameters:
				dir	information flow type
				hashv	hash value
				comm	command
				param	parameter
				ip	ip address of requested data onnection
				port	port of requested data onnection
			Workings:
				accepts it
			Return value:
				Z_ACCEPT
		"""
                return Z_ACCEPT

        def pasvAnswer(self, dir, line, ip, port):
		"""
			ansver to a PASV commend
			Parameters:
				dir	information flow type
				line
				ip	ip address of requested data onnection
				port	port of requested data onnection
			Workings:
				accepts it
			Return value:
				Z_ACCEPT
		"""
                return Z_ACCEPT

	def unknownCommand(self, dir, comm, param):
		"""
			unknown command
			Parameters:
				dir	information flow type
				hashv	hash value
				comm	command
				param	parameter
			Workings:
				accepts it
			Return value:
				Z_ACCEPT
		"""
		return Z_ACCEPT

