############################################################################
##
## Copyright (c) 2000 BalaBit IT Ltd, Budapest, Hungary
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this program; if not, write to the Free Software
## Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
##
## $Id: Proxy.py,v 1.15 2000/10/10 18:25:59 bazsi Exp $
##
## Author  : Bazsi
## Auditor : kisza
## Last version: 1.13
## Notes:
##   - comments in English
##
## Author  : Bazsi
## Auditor : kisza
## Last version: 1.13
## Notes:
##   - comments in English
##
############################################################################

from Zorp import *
import Stream
import new

class Proxy(ZorpProxy):
	"""
		A Proxy-k bzisosztlya

		Vltozk:
			session			A Kapcsolat lerja
		---
		Base class for the proxies
		Attributes:
			session			descriptor of the connection
	"""
	session = None
	
	def __init__(self, name, session):
		"""
			Ltrehozza a proxy objektum egy pldnyt

			Paramterek:
				name			A proxy neve (tipusa)
				session		A kapcsolat lerja

			Lers:
				Belltja az attribtumot, logol, s meghvja az alacsony szint
				objektumot.

			Visszatrsi rtk:
				-
			---
			Constructor
			Parameters:
				name		The type of the proxy
				session		descriptor of the connection
			Workings:
				Sets the attribute, logs, and calls the low-level object
			Return value:
				-
				
		"""
		self.session = session
		session.setProxy(name)
		message(1, "%s: session_start, module=%s" % (session.session_id, name));
		message(1, "%s: client_fd=%d, client_addr=%s" % (session.session_id, session.client_fd, session.client_addr.format()))
		ZorpProxy.__init__(self, name, session.session_id, session.client)


	def addPolicy(self, klass):
		"""
			Adds a policy to the proxy
			Parameters:
				klass		the class implementing the policy
			Workings:
				This is black magic
				Makes klass the parent class of the instance RUNTIME
				Python rulez
			Return value:
				-
		"""
		self.__class__ = new.classobj(self.__class__.__name__, self.__class__.__bases__ + (klass,), self.__class__.__dict__)

	def connectServer(self, host, port):
		"""
			CallBack fuggveny. Akkor hivodik meg, ha krs rkezett a proxy-hoz.

			Paramterek:
				host			A szerver cme.  (HASZNLVA VAN?)
				port			A szerver portja (HASZNLVA VAN?)

			Lers:
				Ha chainer nnics definilva, akkor szerver oldali fd-t probalja
				meg visszaadni, ha ez nem ltezik, hibazenetet ad vissza,
				ha chainer definilva van, akkor azt meghivja a kapott host s
				port paramterekkel.

			Visszatrsi rtk:
				A kapcsolat lerja.
			---
			Callback method called when a connection established
			Parameters:
				host			The address of the server
				port			The port of the server
			Workings:
				If there is no chainer defined, it tries to return the server-side fd
				if it does not exists, raises an error.
				If the chainer is defined, it is called with the given host and port
				paameters.
			Return value:
				The descriptor of the server stream
		"""
		if self.session.chainer == None:
			# we have no chainer, the server side fd should be available
			# by now
			
			if self.session.server_fd == -1:
				raise InternalError, "No chainer and server_fd is -1"
				
			self.session.server = Stream.Stream(self.session.server_fd, self.session.session_id + ": server")
		else:
			self.session.chainer.connectServer(self.session, host, port)
		message(1, "%s: server_fd=%d, server_addr=%s" % (self.session.session_id, self.session.server_fd, self.session.server_addr.format()))
		return self.session.server

class DatagramProxy(ZorpProxy):
	"""
		DataGram orientlt Proxy-k kontnerosztlya

		Vltozk:
			session			A Kapcsolat lerja
		---
		Container class for datagram oriented proxies
		Attributes:
			session		The descriptor of the connection
	"""
	session = None
	
	def __init__(self, name, session):
		"""
			Ltrehozza a proxy objektum egy pldnyt

			Paramterek:
				name			A proxy neve (tipusa)
				session		A kapcsolat lerja

			Lers:
				Logol, hvja az als szintet

			Visszatrsi rtk:
				-
			---
			Creates an instance
			Parameters:
				name		The type of the proxy
				session		The descriptor of the connection
			Workings:
				logs and calls the low-level object
			Return value:
				-
		"""
		self.session = session
		message(1, "%s: session_start, module=%s" % (session.session_id, name));
		ZorpProxy.__init__(self, name, session.session_id, session.client)

