############################################################################
##
## Copyright (c) 2000 BalaBit IT Ltd, Budapest, Hungary
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this program; if not, write to the Free Software
## Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
##
## $Id: Service.py,v 1.11 2000/10/10 18:25:59 bazsi Exp $
##
## Author  : Bazsi
## Auditor : kisza
## Last version: 1.9
## Notes:
##   - comments in English
##
## Author  : Bazsi
## Auditor : kisza
## Last version: 1.9
## Notes:
##   - comments in English
##
############################################################################

import Zorp, Session, Stream

#class ServiceInstance:
#	client = None
#	zone = None
#	
#	def __init__(self, zone, client, param):
#		self.zone = zone
#		self.client = client

class AbstractService:
	"""
		Service kontainer osztlya

		Vltozk:
			name			A service neve
			instance_id	A service ltal kiszolglt session
			Label		The label class used this week
		---
		Container class for the service
		Attributes:
			name		The name of the service
			instance_id	the session serviced by the service
			Label		The label class used this week
					This should be set in initialisation by the administrator

	"""
	name = ""
	instance_id = 0
	
	def __init__(self, name):
		"""
			Ltrehoz egy AbstractService pldnyt.

			Vltozk:
				name		A service neve

			Lers:
				-

			Visszatrsi rtk:
				-
			---
			Creates an instance
			Parameters:
				name		The name of the service
				secattr		The security attributes of the service
			Workings:
				sets the attributes
				handles the case when the label is None (empty label) or
				a tuple which is needed to convert into a Label class
		"""
		self.name = name
	
	def startInstance(self, session):
		"""
			Starts the service
			Parameters:
				session
			Workings:
				not implemented
			Return value:
				-
		"""
		raise NotImplementedError

class Service(AbstractService):
	"""
		Szolgltats osztly feladata a szolgltatshoz tartoz kiegszt
		szolgltatsok elindtsa (esemnykezel)

		Vltozk:
			chainer			A szolgltatshoz tartoz chainer osztly
			proxy_class		A szolgltatshoz tartoz proxy osztlya
			auth		Authentication related data (None if no authentication)
		---
		The Service class starts the complementary services for the
		service (event handler)
		Attributes:
			chainer		The chainer class used for the service
			proxy_class	The proxy class used for the service
			auth		Authentication related data (None if no authentication)
	"""
	chainer = None
	proxy_class = None
	auth = None

	def __init__(self, name, chainer, proxy_class, auth = None):
		"""
			Ltrehozza az osztly egy pldnyt.

			Paramterek:
				chainer		A szolgltatshoz tartoz chainer osztly
				proxy_class	A szolgltatshoz tartoz proxy osztlya

			Lers:
				-

			Visszatrsi rtk:
				-
			---
			Creates an instance
			Parameters:
				chainer		The chainer class used for the service
				proxy_class	The proxy class used for the service
				auth		Authentication related data (None if no authentication)
				label		The security attributes of the service
			Workings:
				sets the attributes
			Return value:
				-
		"""
		AbstractService.__init__(self, name)
		self.chainer = chainer
		self.proxy_class = proxy_class
		self.auth = auth

	def startInstance(self, session):
		"""
			Callback fggvny. A Listener ltal elkapott kapcsolatok
			kezelst vgzi.

			Paramterek:
				session		A kapcsolat lerja (Session.session)

			Lers: 
				Azonostja a felhasznlt, majd ltrehozza a
				kapcsolathoz tartoz callback
				fggvnyosztly egy pldnyt

			Visszatrsi rtk:
				None
			---
			Callback method to handle the new connections
			Parameters:
				session		The description of a connection
			Workings:
				Authenticates the user, and starts a proxy using the
				session derived from the session parameter
				FIXME:	session.auth.authenticate is supposed to raise
					the security attributes maximum to session.max_sec_attr
			Return value:
				None
		"""

		if session.auth == None:
			session.auth = self.auth		
		if session.auth != None and not session.auth.authenticate(session):
			raise Zorp.AuthException

		Zorp.message(0, "auth_user=" + session.auth_user)
		
		session.name = self.name
		instance_id = self.instance_id
		
		# FIXME: handle wraparound
		self.instance_id = self.instance_id + 1
		
		session.setServiceInstance(instance_id)
		ss = Session.StackedSession(session, self.chainer)
		ss.client = Stream.Stream(ss.client_fd, ss.session_id + ": client")
		self.proxy_class(ss)
			
		return None
		
#
# Egy instance_class es param helyett egy mappingot vesz at, ami zonatol
# fuggoen mas es mas eredmenyt ad
#
#class MultiService(AbstractService):
#	def __init__(self, name
