############################################################################
##
## Copyright (c) 2000 BalaBit IT Ltd, Budapest, Hungary
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this program; if not, write to the Free Software
## Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
##
## $Id: Session.py,v 1.11 2000/10/10 18:25:59 bazsi Exp $
##
## Author  : Bazsi
## Auditor : kisza
## Last version: 1.9
## Notes:
##   - comments in English
##
## Author  : Bazsi
## Auditor : kisza
## Last version: 1.9
## Notes:
##   - comments in English
##
############################################################################

from Zorp import *
import os

class AbstractSession:
	"""
	   Session kontnerosztly

		Vltozk:
			session_id			a session azonostja
	"""

	def destroy(self):
		"""
			Kikapcsolja a sessiont

			Paramterek:
				-

			Lers:
				Bezrja a session-hoz tartoz fd-ket
		"""
		if self.client_fd != -1:
			os.close(self.client_fd)
		if self.server_fd != -1:
			os.close(self.server_fd)


class MasterSession(AbstractSession):
	"""
	   Egy alapsession lerja

		Vltozk:
			client*			A klienssel kapcsoaltos vltozk
			server*			A szerver oldallal kapcsoaltos vltozk
				fd				A kapcsolat fd-je
				addr			A kapcsolat tufelnek cme (Sockaddr)
				zone			A kapcsolat Znja
			service			A kapcsolathoz tartoz szolgltats referencija
			instance_id		A kapcsolat egyedi azonostja
	"""

	client = None
	client_fd = -1
	client_addr = None
	client_zone = None

	server = None
	server_fd = -1
	server_addr = None
	server_zone = None
	
	auth = None
	auth_user = ""
	
	service = ""
	instance_id = 0

	session_id = None # "firewall/service:instance id/proxy"
	
	def __init__(self):
		self.session_id = "(%s)" % firewall_name,

	def setService(self, service):
		"""
			Belltja a session-hz tartoz szolgltats.

			Paramterek:
				service		A kapcsolathoz tartoz szolgltats

			Lers:
				Eltrolja, hogy a kapcsolathoz melyik szolgltatst kell
				hasznlni

			Vissztrsi rtk:
				-
		"""
		self.service = service
		self.session_id = "(%s/%s)" % (firewall_name, service.name)

	def setClient(self, fd, addr):
		"""
			Belltja a session kliens rszt.

			Paramterek:
				fd				A kliens fele meno kapcsolat fd-je.
				addr			A kliens cme (SockAddr)

			Lers:
				Eltrolja a klien fele meno kapcsoalt fd-jet, a kliens cmt, 
				s azt, hogy a kliens melyik znba tartozik. Majd megvizsglja,
				hogy a kapcsolat znk alapjn engedlyzehet-e, ennek hinyban
				hibt jelez vissza.

			Visszatrsi rtk:
				-
		"""
		self.client_fd = fd
		self.client_addr = addr
		self.client_zone = findZone(addr)
		if self.client_zone.isOutboundServicePermitted(self) != Z_ACCEPT:
			message(0, "Outbound service %s not permitted from zone %s" % (self.service.name, self.client_zone.name))
			raise DACException

	def setServer(self, addr):
		"""
			Belltja a session szerver rszt.

			Paramterek:
				addr			A szerver cme (SockAddr)

			Lers:
				Eltrolja a szerver cmt, s azt, hogy a kliens melyik znba
				tartozik. Majd megvizsglja, hogy a kapcsolat znk alapjn
				engedlyzehet-e, ennek hinyban hibt jelez vissza.

			Visszatrsi rtk:
				-
		"""
		self.server_addr = addr
		self.server_zone = findZone(addr)
		if self.server_zone.isInboundServicePermitted(self) != Z_ACCEPT:
			message(0, "Inbound service %s not permitted into zone %s" % (self.service.name, self.server_zone.name))
			raise DACException
	
	def setServiceInstance(self, instance_id):
		"""
			Belltja a session pldnynak uniq azonostjt.

			Paramterek:
				instance_id		uniq azonost

			Lers:
				-

			Visszatrsi rtk:
				-
		"""
		self.instance_id = instance_id
		self.session_id = "(%s/%s:%d)" % (firewall_name, self.name, self.instance_id)


class StackedSession(AbstractSession):
	"""
		Alsession osztly

		Vltozk:
			owner				A szl session referencija
			chainer				A session-hoz tartozchainer
	"""
	owner = None	# reference of the owning session
	chainer = None

	def __init__(self, owner, chainer = None):
		self.owner = owner
		self.chainer = chainer
		
	def __getattr__(self, name):
		"""
			A session valtozinak lekrsre alkalmas fggvny

			Paramterek:
				name					A lekrdezett vtoz

			Lers:
				Elsszr megproblja sajt session-jeben megtallni a vltozt, ha
				ez nem sikerl, akkor tadja a krst a szl session-nek.

			Visszatrsi rtk:
				A krt vltoz rtke.
		"""
		try:
			return self.__dict__[name]
		except KeyError:
			return getattr(self.owner, name)

	def setProxy(self, proxy):
		self.session_id = "(%s/%s:%d/%s)" % (firewall_name, self.name, self.instance_id, proxy)
