/***************************************************************************
 *
 * Copyright (c) 2000 BalaBit IT Ltd, Budapest, Hungary
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Id: main.c,v 1.29 2000/10/10 18:25:59 bazsi Exp $
 *
 * Author  : bazsi
 * Auditor : kisza
 * Last version : 1.24
 * Notes   : 
 *   extern void *_proxy_new(); declarations:
 *     needed for sink and anypy, all others: called from the lib
 *
 * Author  : bazsi
 * Auditor : kisza
 * Last version : 1.24
 * Notes   : 
 *   extern void *_proxy_new(); declarations:
 *     needed for sink and anypy, all others: called from the lib
 *
 ***************************************************************************/

#include <zorp/io.h>
#include <zorp/stream.h>
#include <zorp/policy.h>
#include <zorp/registry.h>
#include <zorp/thread.h>
#include <zorp/log.h>
#include <zorp/cfg.h>
#include <zorp/pyauthconn.h>

#include <signal.h>
#include <getopt.h>

extern GList *threads;

void
print_thread(gpointer data, gpointer ignored)
{
  ZThread *thread = (ZThread *) data;
  
  z_message(0, "Thread(%d), name=%s", thread->thread_id, thread->name);
}

gboolean
print_threads(gpointer ignored)
{
  g_message("Thread usage summary:");
  g_list_foreach(threads, print_thread, NULL);
  g_mem_profile();
  return TRUE;
}


GMainLoop *loop;

extern void pop3_proxy_new();
extern void anypy_proxy_new();
extern void sink_proxy_new();
extern void http_proxy_new();

void
sigterm_handler(int signo)
{
  if (loop)
    g_main_quit(loop);
  else
    exit(1);
}

void
z_write_pid_file(char *fname)
{
  FILE *fd = fopen(fname, "w");
  if (fd) 
    {
      fprintf(fd, "%d\n", getpid());
      fclose(fd);
    }
}

void
z_remove_pid_file(char *fname)
{
  unlink(fname);
}

int 
main(int argc, char *argv[])
{
  struct option zorp_options[] = 
  {
    { "as", required_argument, NULL, 'a' },
    { "policy", required_argument, NULL, 'p' },
    { "verbose", optional_argument, NULL, 'v' },
    { "debug", no_argument, NULL, 'd' },
    { "pidfile", required_argument, NULL, 'P' },
    { NULL, 0, NULL, 0 }
  };
  int opt, i;
  gchar *instance_name = "zorp";
  gchar *policy_file = ZORP_POLICY_FILE;
  gchar *auth_cert_file = ZORP_AUTH_CERT_FILE;
  gchar *auth_key_file = ZORP_AUTH_KEY_FILE;
  gchar pid_file[128];

  pid_file[0] = 0;
  while ((opt = getopt_long(argc, argv, "a:p:v::d", zorp_options, NULL)) != -1)
    {
      switch (opt)
	{
	case 'a':
	  instance_name = optarg;
	  break;
	case 'p':
	  policy_file = optarg;
	  break;
	case 'v':
	  if (optarg)
	    verbose_level = atoi(optarg);
	  else
	    verbose_level++;
	  break;
	case 'd':
	  debug_flag++;
	  break;
	case 'P':
	  strncpy(pid_file, optarg, sizeof(pid_file));
	  break;
	default:
	  break;
	}
    }

  if (pid_file[0] == 0)
    {
      g_snprintf(pid_file, sizeof(pid_file), "%s/zorp-%s.pid", ZORP_PID_FILE_DIR, instance_name);
    }
  z_write_pid_file(pid_file);

  signal(SIGPIPE, SIG_IGN);
  g_thread_init(NULL);
  z_log_init(instance_name, 1);
  z_debug(0, "Verbosity level: %d", verbose_level);

  z_registry_init();

  if (!z_python_policy_init())
    {
      z_message(0, "Error initializing Python policy engine.");
      return 1;
    }
  
#if 0
  for (i = 0; i < NSIG; i++)
    siginterrupt(i, 0);
#endif
  signal(SIGINT, sigterm_handler);
  signal(SIGTERM, sigterm_handler); 

  z_config_init();

  z_message(0, "%s version %s starting up", PACKAGE, VERSION);
  
  z_authconn_init(auth_cert_file, auth_key_file);

#if 0
  z_registry_add("plug", ZR_PROXY, plug_proxy_new); 
  z_registry_add("ftp", ZR_PROXY, ftp_proxy_new); 
  z_registry_add("http", ZR_PROXY, http_proxy_new);
#endif

  z_registry_add("sink", ZR_PROXY, sink_proxy_new);
  z_registry_add("anypy", ZR_PYPROXY, anypy_proxy_new); 

/*  z_registry_add("pop3", ZR_PROXY, pop3_proxy_new); */
  
  if (z_policy_read(policy_file))
    {
      z_policy_init(instance_name);
      loop = g_main_new(0);
      /* g_mem_profile(); */
     /* g_timeout_add(10000, print_threads, NULL); */
      g_main_run(loop);
      g_main_destroy(loop);
    }
  else
    {
      z_message(0, "Error opening policy file %s", policy_file);
    }
  z_message(0, "%s version %s going down.", PACKAGE, VERSION);
  z_config_destroy();
  z_policy_destroy();
  z_remove_pid_file(pid_file);
  return 0;
}
